#pragma once


#if defined WIN32 || defined _WINDOWS

#  ifdef BMDS_MODELS_EXPORTS
#    define BMDS_ENTRY_API __declspec(dllexport)
#  else
#    define BMDS_ENTRY_API __declspec(dllimport)
#  endif // BMDS_MODELS_EXPORTS

typedef BSTR BMDS_MODEL_ID;
#else
typedef char * BMDS_MODEL_ID;
#endif // WIN32

static const int NUM_PRIOR_COLS = 5;

#define BMDS_BLANK_VALUE -9999

// The "pack(4)" pragma is required because Excel requires four-byte aligned structures,
// but Visual Studio uses eight-byte boundaries by default.

#pragma pack(show)
#pragma pack(push, 4)
#pragma pack(show)

enum CModelID_t {
  eExp2 = 2,
  eExp3 = 3,
  eExp4 = 4,
  eExp5 = 5,
  eHill = 6,
  ePoly = 7,
  ePow = 8
};

enum BMDSPrior_t {
  eNone = 0,
  eNormal = 1,
  eLognormal = 2
};

enum BMRType_t {
  eAbsoluteDev = 1,
  eStandardDev = 2,
  eRelativeDev = 3,
  ePointEstimate = 4,
  eExtra = 5, // Not used
  eHybrid_Extra = 6,
  eHybrid_Added = 7
};

enum RiskType_t {
  eExtraRisk = 1,
  eAddedRisk = 2
};

enum VarType_t {
  eVarTypeNone = 0,
  eConstant = 1,
  eModeled = 2
};

struct BMDS_C_Options_t {
  double bmr;
  double alpha;
  double background;
  double tailProb;  // Valid only for hybrid bmr type
  int bmrType;
  int degree; // Valid for polynomial type models; for exponential, identifies the submodel
  int adverseDirection; // Direction of adversity: 0=auto, 1=up, -1=down
  int restriction; // Restriction on parameters for certain models
  VarType_t varType;
  bool bLognormal;  // Valid only for continuous models
};

#define NUM_LIKELIHOODS_OF_INTEREST 5
#define NUM_TESTS_OF_INTEREST       4

// Indices for likelihoods of interest
#define LK_A1  0
#define LK_A2  1
#define LK_A3  2
#define LK_R   3
#define LK_FIT 4

// Likelihoods of interest
struct LLRow_t {
  double ll; // Log-likelihood
  double aic;
  int model; // Data model number for test
  int nParms; // Count of model parameters
};

// Indices for tests of interest
#define TI_1 0
#define TI_2 1
#define TI_3 2
#define TI_4 3

// Tests of interest (model deviance tests)
struct TestRow_t {
  double deviance; // -2*log-likelihood ratio
  double pvalue; // test p-value
  int testNumber;
  int df; // test degrees of freedom
};

struct ContinuousDeviance_t {
  LLRow_t *llRows;
  TestRow_t *testRows;
};

struct cGoFRow_t {
  double dose;
  double obsMean;
  double obsStDev;
  double estMean; // Expected dose-response according to the model
  double estStDev;
  double size;
  double scaledResidual;
};

struct cGoF_t {
  double chiSquare;
  double pvalue;
  cGoFRow_t *pzRow;
  int df;
  int n; // Number of rows
};

#pragma pack(show)
struct BMD_C_ANAL {
  BMDS_MODEL_ID model_id;
  double *PARMS;
  ContinuousDeviance_t deviance;
  cGoFRow_t *gofRow; // Goodness of Fit
  bool *boundedParms;
  double MAP;
  double BMD;
  double BMDL;
  double BMDU;
  double AIC;
  double BIC_Equiv; // BIC equivalent for Bayesian runs
  double ll_const; // LL "additive" constant term
  int  nparms;
  bool  bAdverseUp;
};

struct PRIOR {
  double type; // 0= None (frequentist), 1=  normal (Bayesian), 2= log-normal (Bayesian)
  double initalValue;
  double stdDev; // Only used for type= 1 or 2
  double minValue;
  double maxValue;
};

struct PARM_PRIORS {
  int nparms;
  PRIOR *priors;
};

struct MA_PRIORS {
	double *mean_logistic; double *sd_logistic; 
	double *mean_probit;   double *sd_probit  ;
	double *mean_loglogit; double *sd_loglogit; 
	double *mean_logprobit; double *sd_logprobit; 
	double *mean_weibull;  double *sd_weibull; 
	double *mean_gamma;    double *sd_gamma; 
	double *mean_mult2;    double *sd_mult2;
	double *mean_qlinear;  double *sd_qlinear; 
	double *mean_hill;     double *sd_hill; 
};

struct MA_ANALYSIS {
	double BMD; double BMDL; double BMDU; 
	double *bmd;
	double *bmdl; 
	double *bmdu; 
	double *post_probs; 
};

enum BMDSInputType_t {
  unused = 0,
  eCont_2 = 1, // Summarized dose-responses
  eCont_4 = 2, // Individual dose-responses
  eDich_3 = 3, // Regular dichotomous dose-responses
  eDich_4 = 4  // Dichotomous d-r with covariate (e.g., nested)
};

struct BMDSInputData_t {
  double dose;
  double response; // Mean value for summary data
  double groupSize;
  double col4; // stddev for cont_4 or covariate for dich_4
};

#pragma pack(pop)
#pragma pack(show)

extern "C" {

  BMDS_ENTRY_API
    int _stdcall run_cmodel(CModelID_t *p_model, BMD_C_ANAL *returnV, BMDSInputType_t *p_inputType,
    BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDS_C_Options_t *options, int *p_n);
}
