#pragma once

#include "stdafx.h"
#include <chrono>
#include <cmath>
#ifndef WIN32
#include <cfloat>
#endif
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_randist.h>

#include "statmod.h"
#include "dBMDstatmod.h"
#include "cBMDstatmod.h"

#include "log_likelihoods.h"
#include "normal_likelihoods.h"
#include "normalModels.h"
#include "binomModels.h"

#include "IDPrior.h"

#include "DichHillBMD_NC.h"
#include "DichMultistageBMD_NC.h"
#include "DichLogLogisticBMD_NC.h"
#include "DichLogProbitBMD_NC.h"
#include "DichWeibullBMD_NC.h"
#include "DichGammaBMD_NC.h"
#include "DichQlinearBMD_NC.h"
#include "DichLogisticBMD_NC.h"
#include "DichProbitBMD_NC.h"


#include "normal_HILL_NC.h"
#include "normal_POWER_NC.h"
#include "normal_POLYNOMIAL_NC.h"
#include "normal_EXP_NC.h"
#include "lognormal_HILL_NC.h"
#include "lognormal_POWER_NC.h"
#include "lognormal_POLYNOMIAL_NC.h"
#include "lognormal_EXP_NC.h"



class bmd_cdf {
public:
	bmd_cdf():probs(),BMD() {
		spline_bmd_cdf = NULL;
		spline_bmd_inv = NULL;
		acc_bmd_cdf = NULL;
		acc_bmd_inv = NULL;
				
		max_BMD = 0.0; 
		min_BMD = 0.0;

		min_prob = 0.0; 
		max_prob = 0.0; 
	}; 

	bmd_cdf & operator=(const bmd_cdf &M) {
		
		probs = M.probs;
		BMD = M.BMD;

		max_BMD = M.max_BMD; min_BMD = M.min_BMD;
		max_prob = M.max_prob; min_prob = M.min_prob;
		int error = 0;

		if (probs.size() == BMD.size() && BMD.size() >0) {
			acc_bmd_inv = gsl_interp_accel_alloc();
			acc_bmd_cdf = gsl_interp_accel_alloc();
			spline_bmd_inv = gsl_spline_alloc(gsl_interp_steffen, BMD.size());
			spline_bmd_cdf = gsl_spline_alloc(gsl_interp_steffen, BMD.size());
			error = gsl_spline_init(spline_bmd_inv, probs.data(), BMD.data(), BMD.size());
			if (error) {
				spline_bmd_inv = NULL;
				acc_bmd_inv = NULL;
			}

			error = gsl_spline_init(spline_bmd_cdf, BMD.data(), probs.data(), BMD.size());
			if (error) {
				spline_bmd_cdf = NULL;
				acc_bmd_cdf = NULL;
			}
		}
		
		return *this; 
	}
	bmd_cdf(const bmd_cdf &M) {
		*this = M;
	}

	bmd_cdf(std::vector<double> tx, std::vector<double> ty) :probs(tx), BMD(ty) {
		int error; 

		max_prob = *max_element(probs.begin(), probs.end()); 
		min_prob = *min_element(probs.begin(), probs.end());

		max_BMD = *max_element(BMD.begin(), BMD.end());
		min_BMD = *min_element(BMD.begin(), BMD.end());
		
		if (probs.size() == BMD.size() && BMD.size() >0) {
			acc_bmd_inv = gsl_interp_accel_alloc();
			acc_bmd_cdf = gsl_interp_accel_alloc();
			spline_bmd_inv = gsl_spline_alloc(gsl_interp_steffen, BMD.size());
			spline_bmd_cdf = gsl_spline_alloc(gsl_interp_steffen, BMD.size());
			error = gsl_spline_init(spline_bmd_inv, probs.data(), BMD.data(), BMD.size());
		
			if (error) {
				spline_bmd_inv = NULL;
				acc_bmd_inv = NULL;
			}

			error = gsl_spline_init(spline_bmd_cdf, BMD.data(), probs.data(), BMD.size());
			if (error) {
				spline_bmd_cdf = NULL;
				acc_bmd_cdf = NULL;
			}
		}

	}

	~bmd_cdf() {
		if (spline_bmd_inv) {
			gsl_spline_free(spline_bmd_inv);
		}
		if (spline_bmd_cdf) {
			gsl_spline_free(spline_bmd_cdf);
		}
		if (acc_bmd_cdf) {
			gsl_interp_accel_free(acc_bmd_cdf);
		}
		if (acc_bmd_inv) {
			gsl_interp_accel_free(acc_bmd_inv);
		}
		acc_bmd_inv = NULL;
		acc_bmd_cdf = NULL;
		spline_bmd_cdf = NULL; 
		spline_bmd_inv = NULL;
	}

	double inv(double p) {
		if (spline_bmd_cdf !=NULL && acc_bmd_cdf != NULL) {
			if (p > min_prob && p < max_prob)
				return gsl_spline_eval(spline_bmd_inv, p, acc_bmd_inv);
			if (p < min_prob)
				return 0;
			if (p > max_prob)
				return std::numeric_limits<double>::infinity();
		}
		return NAN; 
	}


	double P(double dose) {
		if (spline_bmd_cdf != NULL && acc_bmd_cdf != NULL) {
			if (dose > min_BMD && dose < max_BMD)
				return gsl_spline_eval(spline_bmd_cdf, dose, acc_bmd_cdf);
			if (dose < min_BMD)
				return 0.0;
			if (dose > max_BMD)
				return 1.0; 
		}
		return NAN;
	}
	

private:
	double min_BMD; 
	double max_BMD; 

	double min_prob;
	double max_prob; 
	std::vector<double> probs; 
	std::vector<double> BMD;
	gsl_interp_accel  *acc_bmd_cdf;
	gsl_spline     *spline_bmd_cdf;
	gsl_interp_accel  *acc_bmd_inv;
	gsl_spline     *spline_bmd_inv;
};


struct bmd_analysis {
public:

	bmd_cdf BMD_CDF;
	Eigen::MatrixXd MAP_ESTIMATE;
	Eigen::MatrixXd COV;
	bool isExtra;
	double BMR;
	double MAP_BMD;
	double alpha;
	double MAP;
	contbmd type; 
	std::vector<double> expected; 
	
	bmd_analysis() : MAP_ESTIMATE(), COV(), BMD_CDF() {

	}
	bmd_analysis(const bmd_analysis &M) {
		BMD_CDF = M.BMD_CDF;
		MAP_ESTIMATE = M.MAP_ESTIMATE;
		MAP = M.MAP;
		COV = M.COV;
		BMR = M.BMR;
		MAP_BMD = M.MAP_BMD;
		alpha = M.alpha;
		type = M.type; 
		expected = M.expected; 
	}

	bmd_analysis & operator=(const bmd_analysis &M) {
		BMD_CDF = M.BMD_CDF;
		MAP_ESTIMATE = M.MAP_ESTIMATE;
		MAP = M.MAP;
		COV = M.COV;
		BMR = M.BMR;
		MAP_BMD = M.MAP_BMD;
		alpha = M.alpha;
		type = M.type; 
		expected = M.expected; 
		return *this;
	}

};


Eigen::MatrixXd convertresult_to_probs(Eigen::MatrixXd data) {
	Eigen::MatrixXd result = data;
	bool is_afterMax = false;
	//Transform the profile likelihood into a list of probabilities
	for (int i = 0; i < result.rows(); i++) {
		if (result(i, 2) == 666) {
			is_afterMax = true;
		}
		result(i, 0) = is_afterMax ? .5 + gsl_cdf_chisq_P(2.0*result(i, 0), 1.0) / 2.0 : .5 - gsl_cdf_chisq_P(2.0*result(i, 0), 1.0) / 2.0;

		if (i > 0)
			result(i, 2) = result(i, 0) - result(i - 1);

	}
	// make sure the difference has some small probability of change. 
	for (int i = 1; i < result.rows(); i++) {
		if (result(i, 2) <= 0.0) {
			result(i, 0) = result(i - 1, 0) + 1e-4;
			if (i != result.rows() - 1) {                          // if it is not the last one we have 
				result(i+1, 2) = result(i+1, 0) - result(i,0);    // we must update the difference
			}
		}
	}

	// now we look at the individual rows and make sure there is no nan 
	// if it is a nan we remove that row. 
	int numBad = 0;

	for (int i = 0; i < result.rows(); i++) {
		if (isnan(result(i, 0)) || isinf(result(i, 0))) {
			result(i, 2) = 77;
			numBad++;
		}
	}

	if (numBad) {
		Eigen::MatrixXd tempMat(result.rows() - numBad, result.cols());
		int counter = 0;
		for (int i = 0; i < result.rows(); i++) {
			if (result(i, 2) != 77) {
				tempMat.row(counter) = result.row(i);
				tempMat(counter, 2) = data(i, 2);
				counter++;
			}
		}
		result = tempMat;
	}
	return result;
}



/**********************************************************************
 * function bmd_analysis:
 * 		LL        -    likelihood previously defined
 *      PR        -    Prior
 * 	    fixedB    -    boolean vector of which parameters are fixed
 * 		fixedV    -    vector of what value parameters are fixed too. 
 * 	    BMDType   -    The type of BMD we are searching for
 *      BMRF      -    Benchmark Response
 *      tail_prob -    The tail probability for the Hybrid approach
 *      isIncreasing - true if the BMD should be calculated for an increasing data set
 *      alpha     -    Defines the the lower and upper stopping point for integeration
 *      step_size -    The approximate step size 
 *  returns: 
 *      bmd_analysis class. 
 * *******************************************************************/
template <class LL, class PR>
bmd_analysis bmd_analysis_CNC(LL likelihood, PR prior, 
							  std::vector<bool> fixedB, std::vector<double> fixedV,
							  contbmd BMDType, double BMRF,   double tail_prob, 
							  bool isIncreasing, double alpha, double step_size) {
	// value to return
	bmd_analysis rVal;
	// create the Continuous BMD model
	
	cBMDModel<LL, PR>  model(likelihood, prior, fixedB, fixedV, isIncreasing);								  
	// Find the maximum a-posteriori and compute the BMD
	
	optimizationResult OptRes = findMAP<LL, PR>(&model);
	double BMD = model.returnBMD(BMDType, BMRF, tail_prob); 
	
	Eigen::MatrixXd result = profile_cBMDNC<LL, PR>(  &model,
													  BMDType,
													  BMD,BMRF,tail_prob,
													  step_size,
													  (gsl_cdf_chisq_Pinv(1.0 - 2 * alpha, 1) + 0.1) / 2.0,  // the plus 0.1 is to go slightly beyond
													  isIncreasing);
	

	// Prepare the results to form an X/Y tuple
    // X - BMD values 
    // Y - Cumulative Probabilities associated with the corresponding X row
	//cout << result << endl; 
	result = convertresult_to_probs(result);	
	std::vector<double> x(result.rows());
	std::vector<double> y(result.rows());
	// compute the CDF for the BMD posterior approximation
    if (!isnan(BMD) && BMD > 0 && result.rows() > 5 ) {
		for (int i = 0; i < x.size(); i++) { x[i] = result(i, 0); y[i] = result(i, 1); }
		bmd_cdf cdf(x, y);
		rVal.BMD_CDF = cdf;
	}
	
	Eigen::MatrixXd  estimated = model.log_likelihood.mean(OptRes.max_parms);

	rVal.expected.resize(estimated.rows());

	for (int i = 0; i < rVal.expected.size(); i++) {
		rVal.expected[i] = estimated(i, 0);
	}


	rVal.MAP_BMD = BMD; 
	rVal.BMR = BMRF;
	rVal.isExtra = false;
	rVal.type    = BMDType; 
	rVal.COV = model.varMatrix(OptRes.max_parms);
	rVal.MAP_ESTIMATE = OptRes.max_parms; 
	rVal.MAP = OptRes.functionV; 												   
	
	return rVal; 
									
}

/**********************************************************************
 *  function: bmd_analysis_DNC(Eigen::MatrixXd Y, Eigen::MatrixXd D, Eigen::MatrixXd prior,
 *							 double BMR, bool isExtra, double alpha, double step_size)
 * 
 * 
 * 
 * 
 * 
 * *******************************************************************/
template <class LL, class PR>
bmd_analysis bmd_analysis_DNC(Eigen::MatrixXd Y, Eigen::MatrixXd D, Eigen::MatrixXd prior,
                              std::vector<bool> fixedB, std::vector<double> fixedV, int degree,
                              double BMR, bool isExtra, double alpha, double step_size) {

	LL        dichotimousM(Y, D, degree);
	PR   	  model_prior(prior);

  ofstream file;
  dBMDModel<LL, PR> model(dichotimousM, model_prior, fixedB, fixedV);
  optimizationResult oR = findMAP<LL, PR>(&model);
  bmd_analysis rVal;


	double BMD = isExtra ? model.extra_riskBMDNC(BMR) : model.added_riskBMDNC(BMR);

	//cout << oR.max_parms << endl; 
 // The boolean parameter below specifies robustness - whether to also use MMA opt method   
  Eigen::MatrixXd result = profile_BMDNC<LL, PR>(&model, isExtra,
                          BMD, BMR,
                          step_size, (gsl_cdf_chisq_Pinv(1.0 - 2 * alpha, 1) + 0.1) / 2.0, true);
                          
  
	 
	result = convertresult_to_probs(result);

	//cout << "Result " << result << endl << oR.max_parms << endl; 
	std::vector<double> x(result.rows());
	std::vector<double> y(result.rows());
	for (int i = 0; i < x.size(); i++) { x[i] = result(i, 0); y[i] = result(i, 1); }
	if (!isnan(BMD)) {
		bmd_cdf cdf(x, y);
		rVal.BMD_CDF = cdf;
	}

	Eigen::MatrixXd  estimated_p = model.log_likelihood.mean(oR.max_parms);

	rVal.expected.resize(estimated_p.rows());

	for (int i = 0; i < rVal.expected.size(); i++) {
		rVal.expected[i] = estimated_p(i, 0)*Y(i, 1); 
	}

	rVal.MAP_BMD = BMD; 
	rVal.BMR = BMR;
	rVal.isExtra = isExtra;
	rVal.COV = model.varMatrix(oR.max_parms);
	rVal.MAP_ESTIMATE = oR.max_parms; 
	rVal.MAP = oR.functionV; 
	
	return rVal; 
}


double ma_cdf(double dose, std::vector<double> probs, std::list<bmd_analysis> analyses) {
	double rVal = 0.0;
	double temp;
	int i = 0;
	for (bmd_analysis b : analyses) {
		temp = b.BMD_CDF.P(dose);
		rVal += probs[i] * temp; i++;
	}
	return rVal;
};

double find_maBMDquantile(double q, std::vector<double> probs, std::list<bmd_analysis> analyses) {
	double temp = 0.0;
	double max = 1.0; double min = 0.0;
	bool quantile_isNotBounded = true;
	while (quantile_isNotBounded) {
		temp = ma_cdf(max, probs, analyses);
		if (temp > q) {
			quantile_isNotBounded = false;
		}
		max *= 2;
	}

	double test = (max + min) / 2.0;
	temp = ma_cdf(test, probs, analyses);
	while (abs(log(temp / q)) > 1e-8) { // use the log ratio in the calculation as it is invariant to the size of q
		if (temp > q) {
			max = test;
		}
		else {
			min = test;
		}
		test = (max + min) / 2.0;
		temp = ma_cdf(test, probs, analyses);
	}
	return test;
}
