#include "stdafx.h"
#include <chrono>
#include <iostream>
#include <Eigen/Dense>
#include <iomanip>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_randist.h>

#include "bmd_calculate.h"

#include "statmod.h"
#include "dBMDstatmod.h"
#include "log_likelihoods.h"
#include "binomModels.h"
#include "binomPriors.h"

#include "DichHillBMD_NC.h"
#include "DichMultistageBMD_NC.h"
#include "DichLogLogisticBMD_NC.h"
#include "DichLogProbitBMD_NC.h"
#include "DichWeibullBMD_NC.h"
#include "DichGammaBMD_NC.h"
#include "DichQlinearBMD_NC.h"
#include "DichLogisticBMD_NC.h"
#include "DichProbitBMD_NC.h"
#include "dichotomous_entry.h"

double bmd_log_logistic(
	const  double *d_y, const  double* d_x, const int n) {

	Eigen::MatrixXd  Y(n,1); Eigen::MatrixXd  X(n, 1);

	for (int i = 0; i < n; i++) {
		Y(i, 0) = d_y[i]; X(i, 0) = d_x[i]; 
	}

	Eigen::MatrixXd priors(3, 5);
	priors << 1.0, -2.0, 1.0, -18.0, 18.0,
			  1.0, 0, 1.0, -18, 18.0,
			  2.0, 0, .5, 1e-4, 18.0;

	bmd_analysis a = bmd_analysis_NC<dich_loglogisticModelNC, IDbinomPrior>(Y, X, priors,
																		       0.1, true, 
																			   0.01, 0.05); 

	return a.MAP_BMD;

}

