// dlltemp.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include <fstream>
#include <chrono>
#include <iostream>
#include <Eigen/Dense>
#include <iomanip>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_randist.h>

#include "bmd_calculate.h"

#include "statmod.h"
#include "dBMDstatmod.h"
#include "log_likelihoods.h"
#include "binomModels.h"
#include "binomPriors.h"

#include "DichHillBMD_NC.h"
#include "DichMultistageBMD_NC.h"
#include "DichLogLogisticBMD_NC.h"
#include "DichLogProbitBMD_NC.h"
#include "DichWeibullBMD_NC.h"
#include "DichGammaBMD_NC.h"
#include "DichQlinearBMD_NC.h"
#include "DichLogisticBMD_NC.h"
#include "DichProbitBMD_NC.h"


#include "dichotomous_nc_entry.h"
#include "bmd_calculate.h"


using namespace std; 

#ifndef NUMDOSES
#  define NUMDOSES 4
#endif
#if NUMDOSES == 3
# define INIT_TRIALS {50,50,50}
# define INIT_OBS {1,3,4}
# define INIT_DOSES {0.0, 0.35, 1.0}
# define OBS_SEP_2 '\n'
#elif NUMDOSES == 4
# define INIT_TRIALS {50,50,50,50}
# define INIT_OBS {8,13,18,15}
# define INIT_DOSES {0.0, 0.25, 0.5, 1.0}
# define OBS_SEP_2 ','
# define OBS_SEP_3 '\n'
#else
# define INIT_TRIALS {10,10,10,10,10}
# define INIT_OBS {2,4,7,9,8}
# define INIT_DOSES {0.0, 0.25, 0.5, 0.75, 1.0}
# define OBS_SEP_2 ','
# define OBS_SEP_3 ','
# define OBS_SEP_4 '\n'
#endif

int main (int argc, char *argv[]){
    
	int n = NUMDOSES;
	double trials[NUMDOSES] = INIT_TRIALS;
	double obs[NUMDOSES]    = INIT_OBS;
	double d_x[NUMDOSES]    = INIT_DOSES;
	gsl_set_error_handler_off();
	
	ifstream input(argv[1]); 
	string data; 
	std::string::size_type sz; 
	string string_bmr(argv[2]); 
	if (!input){		
		return -1; // file failed to open 
	}
	
	getline(input,data);  
	for (int i = 0; i < 2000 ; i++){
		std::getline(input,data,',') ;
		std::getline(input,data,',') ;  obs[0] = stod(data,&sz);
		std::getline(input,data,',') ;  obs[1] = stod(data,&sz);
		std::getline(input,data,OBS_SEP_2) ;  obs[2] = stod(data,&sz);
#if NUMDOSES > 3
		std::getline(input,data,OBS_SEP_3);  obs[3] = stod(data,&sz);
#elif NUMDOSES > 4
		std::getline(input,data,OBS_SEP_4);  obs[4] = stod(data,&sz);
#endif
	
 	// Basic Priors		
#ifdef PRISET1
		double mean_logistic[2] = {0.0,0.0}; double sd_logistic[2] =   {2.0,1.0};
		double mean_probit[2] =   {0,0.0}; double sd_probit[2] =     {2.0,1.0}; 
		double mean_qlinear[2] =  {0.0,0.5}; double sd_qlinear[2] =    {2.0,1.0}; 
		
		double mean_loglogit[3] = {0.0,0.0,log(2)}; double sd_loglogit[3] = {2.0,1.0,sqrt(0.25)};
		double mean_logprobit[3] = {0,0.0,log(2)}; double sd_logprobit[3] = {2.0,1.0,sqrt(0.25)}; 
		
		double mean_weibull[3] = {0,log(2),0.0}; double sd_weibull[3] = {2.0,sqrt(0.18),1.0};
		double mean_gamma[3] = {0,log(2),0.0};  double sd_gamma[3] = {2.0,sqrt(0.18),1.0}; 
		
		double mean_mult2[3] = {0.0,0.0,0.0}; double sd_mult2[3] = {2.0,sqrt(0.25),1.0};
		
		double mean_hill[4] = {0.0,4.0,0.0,log(10)}; double sd_hill[4] = {2,2,sqrt(0.25),sqrt(0.06255)};
#endif // PRISET1

		// Diffuse Basic Priors		
#ifdef PRISET2
		double mean_logistic[2] = {0,0.0}; double sd_logistic[2] =   {2.0,2.0};
		double mean_probit[2] =   {0,0.0}; double sd_probit[2] =     {2.0,2.0}; 
		double mean_qlinear[2] =  {0.0,0.5}; double sd_qlinear[2] =    {2.0,2.0}; 
		
		double mean_loglogit[3] = {0.0,0.0,log(2)}; double sd_loglogit[3] = {2.0,1.0,2*sqrt(0.25)};
		double mean_logprobit[3] = {0,0.0,log(2)}; double sd_logprobit[3] = {2.0,1.0,2*sqrt(0.25)}; 
		
		double mean_weibull[3] = {0,log(2),0.0}; double sd_weibull[3] = {2.0,2*sqrt(0.18),2.0};
		double mean_gamma[3] = {0,log(2),0.0};  double sd_gamma[3] = {2.0,2*sqrt(0.18),2.0}; 
		
		double mean_mult2[3] = {0.0,0.0,0.0}; double sd_mult2[3] = {2.0,2*sqrt(0.25),2.0};
		
		double mean_hill[4] = {0.0,4.0,0.0,log(10)}; double sd_hill[4] = {2,2,2*sqrt(0.25),sqrt(2*0.06255)};
#endif // PRISET2

		// Wheeler Suggested Priors 1 
#ifdef PRISET3
		double mean_logistic[2] = {-2.5,0.0}; double sd_logistic[2] =   {1.0,2.0};
		double mean_probit[2] =   {-2.5,0.0}; double sd_probit[2] =     {1.0,2.0}; 
		double mean_qlinear[2] =  {-2.5,0.0}; double sd_qlinear[2] =    {1.0,2.0}; 
		
		double mean_loglogit[3] = {-2.5,0.0,log(2)}; double sd_loglogit[3] = {1.0,1.0,sqrt(0.2)};
		double mean_logprobit[3] = {-2.5,0.0,log(2)}; double sd_logprobit[3] = {1.0,1.0,sqrt(0.2)}; 
		
		double mean_weibull[3] = {-2.5,log(2),0.0}; double sd_weibull[3] = {1.0,sqrt(0.2),2.0};
		double mean_gamma[3] = {-2.5,log(2),0.0};  double sd_gamma[3] = {1.0,sqrt(0.2),2.0}; 
		
		double mean_mult2[3] = {-2.5,0.5,0.0}; double sd_mult2[3] = {1.0,sqrt(0.2),2.0};
		
		double mean_hill[4] = {-2.5,4.0,0.0,log(4)}; double sd_hill[4] = {1,2,sqrt(0.5),sqrt(0.2)};
#endif // PRISET3

		// Wheeler Suggested Priors 2
#ifdef PRISET4
		double mean_logistic[2] = {-2.5,0.0}; double sd_logistic[2] =   {1.0,2.0};
		double mean_probit[2] =   {-2.5,0.0}; double sd_probit[2] =     {1.0,2.0}; 
		double mean_qlinear[2] =  {-2.5,-1.0}; double sd_qlinear[2] =    {1.0,2.0}; 
		
		double mean_loglogit[3] = {-2.5,0.0,log(2)}; double sd_loglogit[3] = {1.0,1.0,sqrt(0.2)};
		double mean_logprobit[3] = {-2.5,0.0,log(2)}; double sd_logprobit[3] = {1.0,1.0,sqrt(0.2)}; 
		
		double mean_weibull[3] = {-2.5,log(2),0.0}; double sd_weibull[3] = {1.0,sqrt(0.2),2.0};
		double mean_gamma[3] = {-2.5,log(2),0.0};  double sd_gamma[3] = {1.0,sqrt(0.2),2.0}; 
		
		double mean_mult2[3] = {-2.5,0.5,0.0}; double sd_mult2[3] = {1.0,sqrt(0.2),2.0};
		
		double mean_hill[4] = {-2.5,0.0,0.0,log(4)}; double sd_hill[4] = {1,2,sqrt(0.5),sqrt(0.2)};
#endif // PRISET4
		
		MA_PRIORS temp; 
		temp.mean_logistic =  mean_logistic; temp.sd_logistic =  sd_logistic; 
		temp.mean_probit =  mean_probit; temp.sd_probit =  sd_probit; 
		temp.mean_qlinear =  mean_qlinear; temp.sd_qlinear =  sd_qlinear; 
		
		temp.mean_logprobit =  mean_logprobit; temp.sd_logprobit = sd_logprobit; 
		temp.mean_loglogit =  mean_loglogit; temp.sd_loglogit =  sd_loglogit; 
		
		temp.mean_weibull =  mean_weibull; temp.sd_weibull =  sd_weibull; 
		temp.mean_gamma =  mean_gamma; temp.sd_gamma =  sd_gamma; 
		
		temp.mean_mult2 =  mean_mult2; temp.sd_mult2 =  sd_mult2; 
		temp.mean_hill =  mean_hill; temp.sd_hill =  sd_hill; 
		
		double bmr = stod(string_bmr,&sz); double alpha = 0.05;
		double p_m_probs[9] = {1.0/9.0,1.0/9.0,1.0/9.0,1.0/9.0,1.0/9.0,1.0/9.0,
								1.0/9.0,1.0/9.0,1.0/9.0};
		double p_m_probs2[9] = {0.0625,0.0625,0.5,0.0625,0.0625,0.0625,
								0.0625,0.0625,0.0625};
	
	     
		double post_p[9]; 
		double ma_bmd[3]; double bmd[9]; double bmdl[9]; double bmdu[9]; 
		double ma_bmd2[3]; 
							
		bmd_MA(obs,trials,d_x,&temp,p_m_probs,&bmr,&alpha,&n,post_p,
				ma_bmd,bmd,bmdl,bmdu); 
		bmd_MA(obs,trials,d_x,&temp,p_m_probs2,&bmr,&alpha,&n,post_p,
			   ma_bmd2,bmd,bmdl,bmdu); 
			   
		cout.setf(ios::fixed, ios::floatfield); 
		cout.precision(8); 
		cout << i << " " << ma_bmd[0] << " " <<  ma_bmd[1] << " " <<  ma_bmd[2] <<  " " << ma_bmd2[0] << " " <<  ma_bmd2[1] << " " <<  ma_bmd2[2] << endl; 
	}
	return 0.0; 
}

void bmd_MA(double *obs, double *trials,
	double *d_x, MA_PRIORS *priors, double *p_m_probs,
	double *bmr, double *alpha, int* n, double * post_p,
	double *ma_bmd, double *bmd,double *bmdl, double *bmdu) {

	gsl_set_error_handler_off();

	Eigen::MatrixXd  Y(*n, 2); Eigen::MatrixXd  X(*n, 1);

	for (int i = 0; i < *n; i++) {
		Y(i, 0) = obs[i]; X(i, 0) = d_x[i];
		Y(i, 1) = trials[i];
	}
	
    // cout << Y << endl; 
	//Logistic Setup
	Eigen::MatrixXd prior_logit(2, 5);
	prior_logit << 1.0, priors->mean_logistic[0], priors->sd_logistic[0],   -20,   20,
		   		   2.0, priors->mean_logistic[1], priors->sd_logistic[1], 1e-12, 40.0;
	IDbinomPrior logistic_prior(prior_logit);

	//Probit Setup
	Eigen::MatrixXd prior_probit(2, 5);
	prior_probit <<  1.0, priors->mean_probit[0], priors->sd_probit[0], -20, 20,
					2.0, priors->mean_probit[1], priors->sd_probit[1], 1e-12, 40;
	IDbinomPrior probit_prior(prior_probit);
	
	//Quantal Linear Setup
	Eigen::MatrixXd prior_qlinear(2, 5);
	prior_qlinear << 1.0, priors->mean_qlinear[0], priors->sd_qlinear[0], -20  , 20,
				   2.0, priors->mean_qlinear[1], priors->sd_qlinear[1], 1e-12, 1000.0;
	IDbinomPrior qlinear_prior(prior_qlinear);

	//Log-Logit Setup
	Eigen::MatrixXd prior_loglogit(3, 5);
	prior_loglogit <<	1.0, priors->mean_loglogit[0], priors->sd_loglogit[0], -20.0, 20.0,
						1.0, priors->mean_loglogit[1], priors->sd_loglogit[1], -40.0, 40.0,
						2.0, priors->mean_loglogit[2], priors->sd_loglogit[2], 1e-4, 40.0;
	IDbinomPrior loglogit_prior(prior_loglogit);

	//Log-Probit Setup
	Eigen::MatrixXd prior_logprobit(3, 5);
	prior_logprobit << 1.0, priors->mean_logprobit[0], priors->sd_logprobit[0], -20.0, 20.0,
					   1.0, priors->mean_logprobit[1], priors->sd_logprobit[1], -40.0, 40.0,
					   2.0, priors->mean_logprobit[2], priors->sd_logprobit[2], 1e-4, 40.0;
	IDbinomPrior logprobit_prior(prior_logprobit);

	//Multistage Setup
	Eigen::MatrixXd prior_mult2(3, 5);
	prior_mult2    << 1.0, priors->mean_mult2[0], priors->sd_mult2[0],-20.0, 20.0,
					  2.0, priors->mean_mult2[1], priors->sd_mult2[1], 1e-4, 1000.0,
					  2.0, priors->mean_mult2[2], priors->sd_mult2[2], 1e-4, 1000.0;
	IDbinomPrior mult2_prior(prior_mult2);


	//Weibull Setup
	Eigen::MatrixXd prior_weibull(3, 5);
	prior_weibull << 1.0, priors->mean_weibull[0], priors->sd_weibull[0], -20.0, 20.0,
					 2.0, priors->mean_weibull[1], priors->sd_weibull[1], 1e-4, 20.0,
					 2.0, priors->mean_weibull[2], priors->sd_weibull[2], 1e-4, 20.0;
	IDbinomPrior weibull_prior(prior_weibull);

	//Gamma Setup
	Eigen::MatrixXd prior_gamma(3, 5);
	prior_gamma   << 1.0, priors->mean_gamma[0], priors->sd_gamma[0], -18.0, 18.0,
					 2.0, priors->mean_gamma[1], priors->sd_gamma[1],  0.2, 20.0,
					 2.0, priors->mean_gamma[2], priors->sd_gamma[2],  1e-4, 20.0;
	IDbinomPrior gamma_prior(prior_gamma);
	
	//Hill Setup
	Eigen::MatrixXd prior_hill(4, 5);
	prior_hill << 1.0, priors->mean_hill[0], priors->sd_hill[0], -40, 40,
				  1.0, priors->mean_hill[1], priors->sd_hill[1], -40, 40,
				  1.0, priors->mean_hill[2], priors->sd_hill[2], -40, 40,
				  2.0, priors->mean_hill[3], priors->sd_hill[3], 1e-8, 40;
	IDbinomPrior    hill_prior(prior_hill);


	ofstream file; 


	std::vector<double> prior_prob(9); for (int i = 0; i < prior_prob.size(); i++) { prior_prob[i] = p_m_probs[i]; }
	std::vector<double> post_prob(9);
	std::list<bmd_analysis> analyses;

	analyses.push_back(bmd_analysis_NC<dich_logisticModelNC, IDbinomPrior>(Y,  X,  prior_logit,  *bmr, true, 0.005, 0.03));
//	cout << " Here 1" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_probitModelNC, IDbinomPrior>(Y , X,    prior_probit, *bmr, true, 0.005, 0.03));
//	cout << " Here 2" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_qlinearModelNC, IDbinomPrior>(Y  , X,  prior_qlinear, *bmr,true, 0.005, 0.03));
//	cout << " Here 3" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_loglogisticModelNC, IDbinomPrior>(Y, X, prior_loglogit, *bmr, true, 0.005, 0.03));
 //  cout << " Here 4" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_logProbitModelNC, IDbinomPrior>(Y, X,   prior_logprobit, *bmr, true, 0.005, 0.03));
//	cout << " Here 5" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_multistageNC<2>, IDbinomPrior>(Y, X,    prior_mult2, *bmr, true, 0.005, 0.03));
//	cout << " Here 6" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_weibullModelNC, IDbinomPrior>(Y, X, prior_weibull, *bmr, true, 0.005, 0.03));
//	cout << " Here 7" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_gammaModelNC, IDbinomPrior>(Y, X, prior_gamma, *bmr, true, 0.005, 0.03));
//	cout << " Here 8" << endl; 
	analyses.push_back(bmd_analysis_NC<dich_hillModelNC, IDbinomPrior>(Y, X, prior_hill, *bmr, true, 0.005, 0.03));
//	cout << " Here 9" << endl; 
	
	////////////////////////////////////////////////////////////////////////////
	// compute the log Laplace Approximation
	int j = 0;
	double v; 
	
	for (bmd_analysis b : analyses) {
		// note if the estimate of the covariance matrix is near singular, the estimate
		// for the determinant may be negative.  In this case we let it be zero
		post_prob[j] = b.MAP_ESTIMATE.rows() / 2 * log(2 * M_PI) - b.MAP + 0.5*log(max(0.0,b.COV.determinant()));// << endl;
		v = post_prob[j]; 
		j++;
	}

	double max_prob = *max_element(post_prob.begin(), post_prob.end());
	double norm_sum = 0.0;
	
	for (int i = 0; i < post_prob.size(); i++) {
		post_prob[i] = post_prob[i] - max_prob + log(prior_prob[i]);
		norm_sum += exp(post_prob[i]);
	}

	for (int i = 0; i < post_prob.size(); i++) {
		post_prob[i] = exp(post_prob[i]) / norm_sum;
		post_p[i] = double(post_prob[i]);
		v = post_p[i]; 
	}

	//returnV->BMD = find_maBMDquantile(0.50, post_prob, analyses);
	float temp;
	temp = float(find_maBMDquantile(alpha[0], post_prob, analyses));
	ma_bmd[1] = double(temp); // find_maBMDquantile(alpha[0], post_prob, analyses);
	temp = float(find_maBMDquantile(1.0 - alpha[0], post_prob, analyses));
	ma_bmd[2] = double(temp); // find_maBMDquantile(1.0 - alpha[0], post_prob, analyses);
	int k = 0; 
	ma_bmd[0] = 0.0; 
	// individual BMD analysis

	for (bmd_analysis n : analyses) {
		temp = (float(n.MAP_BMD)); 
		bmd[k] = double(temp);
		temp = float(post_prob[k] * n.MAP_BMD);
		ma_bmd[0] += double(temp);
		temp = float(n.BMD_CDF.inv(alpha[0]));
		bmdl[k] = double(temp);
		temp = float(n.BMD_CDF.inv(1.0 - alpha[0]));
		bmdu[k] = double(temp); // n.BMD_CDF.inv(1.0 - alpha[0]);
		k++; 
	}
	
	k = 0; 
	
	

	

}
