#pragma once

#if defined WIN32 || defined _WINDOWS

#  ifdef BMDS_MODELS_EXPORTS
#    define BMDS_ENTRY_API __declspec(dllexport)
#  else
#    define BMDS_ENTRY_API __declspec(dllimport)
#  endif // BMDS_MODELS_EXPORTS

//typedef BSTR BMDS_MODEL_ID;
typedef char * BMDS_MODEL_ID;
#else
typedef char * BMDS_MODEL_ID;
#endif // WIN32

#define BMDS_BLANK_VALUE -9999

// The "pack(4)" pragma is required because Excel requires four-byte aligned structures,
// but Visual Studio uses eight-byte boundaries by default.
#pragma pack(4)

enum RiskType_t {
  eExtraRisk = 1,
  eAddedRisk = 2
};

struct MSCombo_Options {
  int bRestrict;
  int bmrType; // risk type
  double bmr; // Effect specified by user
  double alpha; // 1 - Confidence Level
  // background and degree are associated with the data sets for ms_combo
};
typedef struct MSCombo_Options MSCombo_Options_t;

struct BMD_ANAL {
  BMDS_MODEL_ID model_id; 
	double MAP; 
	double BMD; 
	double BMDL;
	double BMDU; 
	double *PARMS; 
	int  nparms; 
};
typedef struct BMD_ANAL BMD_ANAL_t;

struct MSCombo_Result {
  BMDS_MODEL_ID model_id;
  double MAP;
  double BMD;
  double BMDL;
  double BMDU;
  double combinedLL; /* Combined log-likelihood */
  double combinedLLConstant; /* not a clue */
  BMD_ANAL_t *dsResults;	/* Results for each individual data set / tumor */
};
typedef struct MSCombo_Result MSCombo_Result_t;

enum BMDSInputType_t {
  unused = 0,
  eCont_2 = 1, // Summarized dose-responses
  eCont_4 = 2, // Individual dose-responses
  eDich_3 = 3, // Regular dichotomous dose-responses
  eDich_4 = 4  // Dichotomous d-r with covariate (e.g., nested)
};

// This is identical to the BMDSInputData_t type
struct BMDSDataRow {
  double dose;
  double response; // Mean value for summary data
  double groupSize;
  double col4; // stddev for cont_4 or covariate for dich_4
};
typedef struct BMDSDataRow BMDSDataRow_t;

struct BMDSDataSet {
  BMDSDataRow_t *rows;		/* Dose-response rows */
  int count;			/* Number of rows */
  int dgCount;			/* Number of dose groups */
  int degree;			/* Polynomial degree to use (if applicable) */
  double background;		/* Assumed response without exposure */
};
typedef struct BMDSDataSet BMDSDataSet_t;

#pragma pack()

#ifdef __cplusplus
extern "C" {
#endif

  BMDS_ENTRY_API
    int _stdcall ms_combo(MSCombo_Result_t *zOut,
      int dsCount,
      BMDSDataSet_t *dataSets,
      MSCombo_Options_t *options);
#ifdef __cplusplus
} /* end extern "C" */
#endif
