#pragma once

#ifdef WIN32

#  ifdef BMDS_MODELS_EXPORTS
#    define BMDS_ENTRY_API __declspec(dllexport)
#  else
#    define BMDS_ENTRY_API __declspec(dllimport)
#  endif // BMDS_MODELS_EXPORTS

//typedef BSTR BMDS_MODEL_ID;
typedef char * BMDS_MODEL_ID;
#else
typedef char * BMDS_MODEL_ID;
#endif // WIN32

#define BMDS_BLANK_VALUE -9999

// The "pack(4)" pragma is required because Excel requires four-byte aligned structures,
// but Visual Studio uses eight-byte boundaries by default.
#pragma pack(4)

enum RiskType_t {
  eExtraRisk = 0,
  eAddedRisk = 1
};

enum LSCType {
  eOverallMean = 0,
  eControlGroupMean = 1
};
typedef enum LSCType LSCType_t;

enum ILCCalc {
  eAssumeZero = 0,
  eEstimate = 1
};
typedef enum ILCCalc ILCCalc_t;

struct NDichotomous_Options {
  int bRestrict; // 0=do not restrict rho parameter, 1=restrict rho >= 1
  int bmrType; // risk type
  double bmr;  // Effect specified by user
  double alpha; // 1 - Confidence Level
  double background;
  int useLSC;  // 0=Do not use LSCs 1=Use LSC
  int LSCType; // For LSC, 0=use overall mean, 1=use cntrl. grp. mean
  int ILCCalc; // For ILC, 0=assume phi parms = 0, 1=estimate them
  int BSSeed;  // Seed value for bootstrapping (LCO: THIS MIGHT NEED TO BE A LONG!!!)
  int BSIter;  // Number of bootstraps for nested p values
};
typedef struct NDichotomous_Options NDichotomous_Options_t;

struct BMD_ANAL {
  BMDS_MODEL_ID model_id; 
	double MAP; 
	double BMD; 
	double BMDL;
	double BMDU; 
	double *PARMS; 
	int  nparms; 
};
typedef struct BMD_ANAL BMD_ANAL_t;

struct NDLitterRow {
  double dose;
  double lsc;
  double estProb;
  double litterSize;
  double expected;
  double observed;
  double scaledResidual;
};
typedef struct NDLitterRow NDLitterRow_t;

struct NDPlotRow {
  double dose;
  double response; // Composite response across litters
  double ebLower;  // Error bar lower limit
  double ebUpper;  // Error bar upper limit
};
typedef struct NDPlotRow NDPlotRow_t; // Composite ("reduced") responses for plotting

struct NDBootstrapRow {
  int bsNum;
  double pvalue;
  double pct50;
  double pct90;
  double pct95;
  double pct99;
};
typedef struct NDBootstrapRow NDBootstrapRow_t;

struct NDichotomous_Result {
  BMDS_MODEL_ID model_id;
  double AIC;
  double MAP; // MAP==MLL for frequentist models
  double BMD;
  double BMDL;
  double BMDU;
  double *PARMS;
  int  nparms;
  int BSSeed; // Bootstrap seed value (LCO: THIS MIGHT NEED TO BE A LONG!!!)
  NDLitterRow_t *litterRow;
  NDBootstrapRow_t *bsRow; // chi^2 bootstrap rows
  NDPlotRow_t *plotData;   // Composite response data for plotting
  double obsChi2;
  double df;
  double srMin;
  double srAvg;
  double srMax;
  double fixedLSC;
  int srNumLitters;
};
typedef struct NDichotomous_Result NDichotomous_Result_t;

enum BMDSInputType_t {
  unused = 0,
  eCont_2 = 1, // Summarized dose-responses
  eCont_4 = 2, // Individual dose-responses
  eDich_3 = 3, // Regular dichotomous dose-responses
  eDich_4 = 4  // Dichotomous d-r with covariate (e.g., nested)
};

// This is identical to the BMDSInputData_t type
struct BMDSDataRow {
  double dose;
  double response; // Mean value for summary data
  double groupSize;
  double col4; // stddev for cont_4 or covariate for dich_4
};
typedef struct BMDSDataRow BMDSDataRow_t;

struct BMDSDataSet {
  BMDSDataRow_t *rows;		/* Dose-response rows */
  int count;			/* Number of rows */
  int dgCount;			/* Number of dose groups */
  int degree;			/* Polynomial degree to use (if applicable) */
  double background;		/* Only used for ms_combo */
};
typedef struct BMDSDataSet BMDSDataSet_t;

#pragma pack()

#ifdef __cplusplus
extern "C" {
#endif

  BMDS_ENTRY_API
    int _stdcall nlogistic(NDichotomous_Result_t *zOut,
      int dsCount,
      BMDSDataSet_t *dataSets,
      NDichotomous_Options_t *options);
#ifdef __cplusplus
} /* end extern "C" */
#endif
