#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <float.h>
//#include <benchmark.h>
//#include <allo_memo.h>
#include "specialfun.h"
#include "bmds_nested_entry.h"
#include "gsl/gsl_randist.h"

/* This is copied from benchmark.h so we don't need to include it */
/*data type for ANOVA summary*/              
typedef struct ana_struct {
	double SS;
	double MSE;
	int    DF;
	double TEST;
} AnaList;

#include "nlogistic_dll.h"

int *IVECTOR();
void  FREE_IVECTOR ();
double *DVECTOR();
void FREE_DVECTOR ();
double **DMATRIX();
void FREE_DMATRIX ();
int compare_doubles (const void *a, const void *b);
double gsl_cdf_beta_P (double x, double a, double b);
void SortByLs (int, int, int [], double [], double [], double [], double [], double []);

extern FILE *fp_out;

//extern int Model_DF (int []);
//extern void Predict(double [], double [], int, double [], double []);

void
myGoodness (NDichotomous_Result_t *zOut, int ngrp, int nparm, double Parms[],
	    int bounded[], int Nobs, double Xi[], double Yp[], double Yn[],
	    double Ls[], int Xg[], double SR[] )
{
  NDLitterRow_t *litterRow = zOut->litterRow;
  double *GYpp;			/*predicted positive depdendent values */
  double *GEp;			/*estimated probability */
  double *P;
  double *Gvar, *GYsum, *Gcount;
  int *GrpSize;			/*numbers of obs. in each group. */
  int i, j, k, l;
  double gfit;
    
  /* Note: have to redefine those var. because the data must be */
  /* changed in order to compute goodness-of-fit statistic. */
  GYpp = DVECTOR (1, Nobs);
  GEp = DVECTOR (1, Nobs);
  GYsum = DVECTOR(1, Nobs);
  Gcount = DVECTOR(1, Nobs);
  Gvar = DVECTOR(1, Nobs);
  GrpSize = IVECTOR (1, ngrp);
  P = DVECTOR (1, nparm);

  /**** compute the GrpSize[] -- # of obs. in each group***/
  for (i = 1; i <= ngrp; i++) GrpSize[i] = 0;
  for (j = 1; j <= Nobs; j++) GrpSize[Xg[j]] += 1;
  for (j = 6; j <= nparm; j++) P[j - 5] = Parms[j]; 

  /** compute the estimated prob. and expected obs.  *******/

  Predict(Xi, Ls, Nobs, Parms, GEp);
  SortByLs (Nobs, ngrp, GrpSize, Ls, Yp, Yn, GYpp, GEp);

  for (i = 1; i <= Nobs; i++) {
    GYsum[i] = Yp[i] + Yn[i];
    GYpp[i] = GEp[i] * GYsum[i];
    Gvar[i] = GYsum[i] * GEp[i] * (1 - GEp[i]) * (1.0 + (GYsum[i] - 1.0) * P[Xg[i]]);
  }

  /*** compute the goodness-of-fit test for litter data. *********/
  gfit = 0.0;
  for (k = 1; k <= Nobs; k++) {
    if (Gvar[k] > 0) {
      SR[k] = Gvar[k] > 0 ? (Yp[k] - GYpp[k])/sqrt(Gvar[k]) : 0;
      gfit += SR[k]*SR[k];
    }
  }
  i = 0;
  for (j = 0; j < Nobs; j++) {
    i++;
    litterRow[j].dose = Xi[i];
    litterRow[j].lsc = Ls[i];
    litterRow[j].estProb = GEp[i];
    litterRow[j].litterSize = GYsum[i];
    litterRow[j].expected = GYpp[i];
    litterRow[j].observed = Yp[i];
    litterRow[j].scaledResidual = (Gvar[i] > 0 ? (Yp[i] - GYpp[i])/sqrt(Gvar[i]) : 0);
  }

  FREE_DVECTOR (GYpp, 1, Nobs);
  FREE_DVECTOR (GEp, 1, Nobs);
  FREE_DVECTOR (GYsum, 1, Nobs);
  FREE_DVECTOR (Gcount, 1, Nobs);
  FREE_DVECTOR (Gvar, 1, Nobs);
  FREE_IVECTOR (GrpSize, 1, ngrp);
  FREE_IVECTOR (P, 1, nparm);
} // end myGoodness

void myPrintText(FILE *fp, const char *pText)
{
  fprintf(fp, "%s\n", pText);
  fflush(fp);
  return;
}

void myPrintInitParms(FILE *fp, int nparm, int Spec[], double Parms[], char *tparms[])
{
  int i;
  for (i = 1; i <= nparm; i++) {
    if (Spec[i] == 0) fprintf (fp, "%31s = %12.6g\n" , tparms[i - 1], Parms[i]);
    else fprintf (fp, "%31s = %12.6g   Specified\n" , tparms[i - 1], Parms[i]);
  } /* end for */
  fflush(fp);
  return;
} // printInitParms()

void myErrorPrt(FILE *fp, const char *text)
{
  fprintf(stderr, "\n%s\n", text);
  fprintf(fp, "\n%s", text);
  fflush(fp);
  //exit (1);                                
}

void myWarning(FILE *fp, const char *text)
{
  fprintf(fp, "\n%s\n", text);
  fflush(fp);
  return;
}

void myOUTPUT_DTMS3PARMS (int nparm, int Spec[], int bounded[], double Parms[],
			  const char *tparms[], double **vcv, int print_SE)
{
  int i, jvar;

  /*output to *.out file */
  myPrintText(fp_out, "\n\n\n                                Parameter Estimates");
  myPrintText(fp_out,     "\n       Variable           Estimate             Std. Err. ");
  jvar = 0;
  for (i = 1; i <= nparm; i++) {
    if(print_SE > 0) {
      if (Spec[i] == 0) {
	if (bounded[i] == 0) {
	  jvar++;
	  if (vcv[jvar][jvar] > 0) {
	    fprintf (fp_out, 
		     "%15s %19.6g %19.6g\n"
		     , tparms[i - 1], Parms[i],
		     sqrt (vcv[jvar][jvar]));
	  }
	  else {
	    fprintf (fp_out, 
		     "%15s %19.6g %19s\n"
		     , tparms[i - 1], Parms[i],
		     "NA");
	  }
	}
	else {
	  fprintf (fp_out, 
		   "%15s %19.6g             Bounded\n"
		   , tparms[i - 1], Parms[i]);
	}
      }
    }
    else { //print_SE == 0
      if (Spec[i] == 0) {
	if (bounded[i] == 0) {
	  jvar++;
	  fprintf (fp_out, 
		   "%15s %19.6g %13s\n"
		   , tparms[i - 1], Parms[i],
		   "*");
	}
	else {
	  fprintf (fp_out, 
		   "%15s %19.6g             *\n"
		   , tparms[i - 1], Parms[i]);
	}
      }
    }
  } //end of for(i = 1; i <= nparm; i++)
  if(print_SE == 0) /* Don't Print Standard Error */
    fprintf (fp_out, "\n* - Indicates that this value is not calculated.\n");
} // end myOUTPUT_DTMS3PARMS()

/*********************************************************************************
 *
 *  OUTPUT_DTMS3VCV--This function will take the matrix of Likelihood
 *		    second derivatives and remove the columns/rows
 *		    that correspond to any bounded or fixed parameters,
 *		    invert this modified matrix, then output it.
 *
 *  INPUT:
 *      nparm       number of parameters in the model
 *      Spec[i]     0 if parameter i is not specified by user
 *		   1 if parameter i is specified by user
 *      parmtxt[i]  the ith parameters name
 *      vcv[][]     contains the second partial derivatives of
 *	           the likelihood function.  Element ij is
 *		   the second partial with respect to
 *		   parameters i and j
 *      bounded[i]  0 if the parameter has not been estimated
 *		   at a boundary point
 *		   1 if the parameter has been estimated at a
 *		   boundary point
 *  INPUT/OUTPUT:
 *	vcv_adj[][]-	Upon entry, is just an uninitialized matrix
 *			of size MxM where M = # of non-bounded
 *			parameters.  At return, it is the corr.
 *			matrix with all bounded and specified
 *			parameters removed
 **********************************************************************************/
void myGet_and_OUTPUT_DTMSVCV (int nparm, int Spec[], char *parmtxt[],
			       double **vcv, double **vcv_adj, int *bounded)
{
  int adj_vcv_rows, i, j, checkvar, counter, *notbounded, nboundparm;
  double cor, **temp_vcv;
  double **DMATRIX(); void FREE_DMATRIX ();
  int *IVECTOR(); void  FREE_IVECTOR ();
  int Take_Out_Bounded_Parms(int, int *, double **);
  int INVMAT(double **, int);
  
  /* Create an adjusted variance-covariance matrix that does not have
     correlations/variances for parameters that are estimated at a boundary point */

  temp_vcv = DMATRIX (1, nparm, 1, nparm);

  for (i = 1; i <= nparm; i++) {
    for (j = 1; j <= nparm; j++) {
      temp_vcv[i][j] = vcv[i][j];
    }
  }


  nboundparm = 0;
  for (i = 1; i <= nparm; i++) {
    if (bounded[i] == 1) {
      nboundparm++;
    }			/* end if */
  }				/* end for */

  notbounded = IVECTOR (1, nparm - nboundparm);
  counter = 0;

  for (i = 1; i <= nparm; i++) {
    if (bounded[i] == 0) {
      /* notbounded[j] contains the subscript in bounded that contains the
	 jth non-bounded parameter */

      counter++;
      notbounded[counter] = i;
    }			/* end if */
  }				/* end for */

  /* Adjust matrix to only contain non-fixed or bounded parameters */
  adj_vcv_rows = Take_Out_Bounded_Parms (nparm, bounded, vcv);

  for (i = 1; i <= adj_vcv_rows; i++) {
    for (j = 1; j <= adj_vcv_rows; j++) {
      vcv_adj[i][j] = vcv[i][j];
    }			/* end for j */
  }				/* end for i */


  /* Invert to get the covariance matrix */

  INVMAT (vcv_adj, adj_vcv_rows);

  /*output to screen and bmdswrk.002 temp file */
  myPrintText(fp_out, "\n\n           Asymptotic Correlation Matrix of Parameter Estimates\n");

  if (nboundparm > 0) {
    /* printf ("           ( *** The model parameter(s) "); */
    fprintf (fp_out, "           ( *** The model parameter(s) ");
    for (i = 1; i <= nparm; i++) {
      if (bounded[i] == 1) {
	/* printf (" -%s   ", parmtxt[i - 1]); */
	fprintf (fp_out, " -%s   ", parmtxt[i - 1]);
      }
    }
    fprintf (fp_out, "\n                 have been estimated at a boundary point, or have been specified by the user,");
    fprintf (fp_out, "\n                 and do not appear in the correlation matrix )\n\n");
  }

  fprintf (fp_out, "          ");

  counter = 0;

  for (i = 1; i <= nparm; i++) {
    if (bounded[i] == 0) {
      /* printf ("%15s", parmtxt[i - 1]); */
      fprintf (fp_out, "%13s", parmtxt[i - 1]);
    }
  }
  fprintf (fp_out, "\n");


  for (i = 1; i <= adj_vcv_rows; i++) {
    fprintf (fp_out, "\n%10s", parmtxt[notbounded[i] - 1]);
    for (j = 1; j <= adj_vcv_rows; j++) {
      if (vcv_adj[i][i] <= 0.0 || vcv_adj[j][j] <= 0.0) {
	fprintf (fp_out, "      NA       ");
	checkvar = 1;
      }
      else {
	cor = vcv_adj[i][j] /
	  (sqrt (vcv_adj[i][i]) * sqrt (vcv_adj[j][j]));
	fprintf (fp_out, 
		 "%13.2g", 
		 cor);
	checkvar = 0;
      }
    }
    fprintf (fp_out, "\n");
  }

  if (checkvar == 1) {
    fprintf (fp_out,
	     "\n\nNA - This parameter's variance has been estimated as zero or less.\nTHE MODEL HAS PROBABLY NOT CONVERGED!!!\n");
  }

  for (i = 1; i <= nparm; i++) {
    for (j = 1; j <= nparm; j++) {
      vcv[i][j] = temp_vcv[i][j];
    }
  }
    
  FREE_IVECTOR(notbounded,1,nparm - nboundparm);
  FREE_DMATRIX(temp_vcv,1,nparm,1,nparm);
} // myGet_and_OUTPUT_DTMSVCV()

void myOUTPUT_DTMS3ANOVA (char *anatxt[], AnaList anasum[])

{
  int Nparms;

  /*output to screen */
  myPrintText(fp_out, "\n\n\n                        Analysis of Deviance Table");
  myPrintText(fp_out, "\n       Model      Log(likelihood)  # Param's  Deviance  Test d.f.   P-value");

  /*output to *.out file */
  Nparms = anasum[3].DF + 1; /* for full model == # dose groups */
  fprintf (fp_out, 
	   "%15s %15.6g %9d\n"
	   , anatxt[0], anasum[1].SS, Nparms);
  Nparms = Nparms - anasum[2].DF; /* # fitted parameters in fitted model */
  if (anasum[2].DF > 0) {
    if (anasum[2].TEST > .0001) {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d %15.4g\n"
	       , anatxt[1], anasum[2].SS, Nparms, anasum[2].MSE, anasum[2].DF,
	       anasum[2].TEST);
    }
    else {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d %19.8g\n"
	       , anatxt[1], anasum[2].SS, Nparms, anasum[2].MSE, anasum[2].DF,
	       anasum[2].TEST);
    }
  }
  else {
    fprintf (fp_out, 
	     "%15s %15.6g %9d %13.6g %6d         NA\n"
	     , anatxt[1], anasum[2].SS, Nparms, anasum[2].MSE, anasum[2].DF);
  }
  Nparms = 1; /* reduced model has just 1 parameter */
  if (anasum[1].DF > 1) {
    if (anasum[3].TEST > .0001) {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d %15.4g\n"
	       , anatxt[2], anasum[3].SS, Nparms, anasum[3].MSE, anasum[3].DF,
	       anasum[3].TEST);
    }
    else {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d         <.0001\n"
	       , anatxt[2], anasum[3].SS, Nparms, anasum[3].MSE, anasum[3].DF);
    }
  }
  else {
    fprintf (fp_out, 
	     "%15s %15.6g %9d %13.6g %6d         NA\n"
	     , anatxt[2], anasum[3].SS, Nparms, anasum[3].MSE, anasum[3].DF);
  }

  fprintf (fp_out,
	   "\n%15s %15.6g\n"
	   , "AIC:",
	   -2*anasum[2].SS + 2*(1.0 + anasum[3].DF - anasum[2].DF));

} // myOUTPUT_DTMS3ANOVA()

void my_do_dmngb_warning(FILE *fp, int * eflag)
{
  if (*eflag > 0) myWarning(fp, "-- NOTE ON MODEL CONVERGENCE:\n");
  if (*eflag == 9)
    {
      myWarning(fp, "  The allowed number of iterations has been reached with no convergence.\n\
  Try restarting and allowing more iterations.");
      myWarning(fp, "  Sometimes this is due to too-stringent convergence criteria.");
      myWarning(fp, "  If the graph indicates a good fit, try increasing the relative");
      myWarning(fp, "  function convergence criterion, for example, from 2e-16 to");
      myWarning(fp, "  2e-08, or increasing the parameter convergence parameter.");
      myWarning(fp, "\n  The tables below indicate the parameter and likelihood");
      myWarning(fp, "  values that correspond to the graph.");
    } /* end if */
  if (*eflag == -1)
    myWarning(fp, "-- All parameters have been specified.\n");
  if (*eflag != 0 && *eflag != 9 && *eflag != -1 && *eflag != 7)
    {
      fprintf(fp, "\n\n  ErrorFlag is: %d\n",*eflag);
      myWarning(fp, "  Error: model failed to converge.\n\
Try restarting with new initial values.\n");
      myWarning(fp, "\n  Tip: Go into Advanced User Mode (on the Options Menu)");
      myWarning(fp, "  of BMDS).  Specify all the parameter values to be the same");
      myWarning(fp, "  as the initial values listed in the output from this run.");
      myWarning(fp, "  Rerun the model, and look at the graph.  Then, repeatedly");
      myWarning(fp, "  adjust the parameters (mostly slope and power) until the");
      myWarning(fp, "  curve comes close to the data.  Then, in the next try,");
      myWarning(fp, "  switch all the \"specified\" labels to \"initialized\", and");
      myWarning(fp, "  rerun the model.");
    }
  if (*eflag == 7)
    {
      myWarning(fp, "  The model appears to have converged, in the sense that a");
      myWarning(fp, "  reasonable change in the parameters is unlikely to increase");
      myWarning(fp, "  the likelihood any further.  However, the hessian matrix at");
      myWarning(fp, "  the reported solution appears to be singular, so this may not");
      myWarning(fp, "  be a true maximum.");
      myWarning(fp, "     Try restarting from different initial values.  Also,");
      myWarning(fp, "  check the goodness of fit table and the graph of the function");
      myWarning(fp, "  before accepting the results of this analysis.  This result can");
      myWarning(fp, "  happen if the parameters are not uniquely determined, so that,");
      myWarning(fp, "  for example, a continuum of values for (Slope, Power) in the Gamma");
      myWarning(fp, "  model yield the same likelihood value.\n");
      *eflag = 0; /* So everything else will work */
    }
}

void
myBootstrap (NDichotomous_Result_t *zOut, int ngrp, int nparm, double Parms[], int bounded[],
	    int Nobs, double Xi[], double Yp[], double Yn[], double Ls[],
	    int Xg[], double SR[], int BSIter, long BSSeed )
{
  double *GYpp;			/*predicted positive depdendent values */
  double *GEp;			/*estimated probability */
  double *P;
  double *Gvar, *GYsum, *Gcount;
  int *GrpSize;			/*numbers of obs. in each group. */
  int i, j, k, l;
  int df;               /*degrees of freedom*/
  double gfit;
  NDBootstrapRow_t *bsRow = zOut->bsRow;
  
  /*** bootstrap variables */
  const gsl_rng_type * type;    /* type for beta distribution */
  gsl_rng * r;                  /* random value for beta distribution */
  double urand;                 /* uniform random value [0,1]  */
  double a, b;                  /* values used in beta distribution function */
  double cutpoint;              /* cut-off probability for bootstrapping */
  double *GYp_new;              /* psuedo-observed values */
  double *GSR_new;              /* SR values based on new variables */
  double **GSR_newsqsum;	/* Sum of SR^2 values for new method */
  int SRsqCount;                /* tracks SR^2 >= SR^2_original */
  int BSLoops = 3;              /* number of times to repeat bootstrap method */
  double *pv;
  double pavg;
  double percentiles[] = {.50, .90, .95, .99};	/* percentile values to calculate */
  int perloc;
  double x, cum_beta, cum_beta_comp, cum_beta_comp_low, cum_beta_comp_hi;  /*cumulative beta variables for NaN method */
  double cum_beta_step = 0.00001;   /* step size for cumulative beta table */
  double eps = 10e-8;  /*accuracy to test for est.prob = 0 */
  double pctTemp[4]; // Used for copying values to output structure
  /*** end of new method variables */
    

  /* Note: have to redefine those var. because the data must be */
  /* changed in order to compute goodness-of-fit statistic. */
  GYpp = DVECTOR (1, Nobs);
  GEp = DVECTOR (1, Nobs);
  GYsum = DVECTOR(1, Nobs);
  Gcount = DVECTOR(1, Nobs);
  Gvar = DVECTOR(1, Nobs);
  GrpSize = IVECTOR (1, ngrp);
  P = DVECTOR (1, nparm);

  /*** for bootstrap method */
  GYp_new = DVECTOR (1, Nobs);  
  GSR_new = DVECTOR (1, Nobs);
  GSR_newsqsum  = DMATRIX (1, BSLoops, 1, BSIter);
  pv = DVECTOR (1,BSLoops);	
#ifdef BSDEBUG
  double *GSR;			/* Add capablility of storing SR values for debug output */   
  FILE *fp_bsdbg;  
  fp_bsdbg = fopen("BSDEBUG.log", "w"); 
  if (fp_bsdbg == NULL) 
    {
    printf ("Error in opening BSDEBUG file\n"); 
    printf ("...now exiting to system...\n"); 
    exit(1);
    }
#endif



/**** compute the GrpSize[] -- # of obs. in each group***/
  for (i = 1; i <= ngrp; i++)
    GrpSize[i] = 0;

  for (j = 1; j <= Nobs; j++)
    GrpSize[Xg[j]] += 1;


  for (j = 6; j <= nparm; j++)
    P[j - 5] = Parms[j]; 

/** compute the estimated prob. and expected obs.  *******/
  Predict(Xi, Ls, Nobs, Parms, GEp);

  

  for (i = 1; i <= Nobs; i++)
    {
      GYsum[i] = Yp[i] + Yn[i];
      GYpp[i] = GEp[i] * GYsum[i];
      Gvar[i] = GYsum[i] * GEp[i] * (1 - GEp[i]) * (1.0 + (GYsum[i] - 1.0) * P[Xg[i]]);
    }


#ifdef BSDEBUG
  /*** Add capability to tally # of observed at each value */
  int max = GYsum[1];
  for (i = 2; i <= Nobs; i++)
    {
      if (max < (int)GYsum[i])
        max = (int)GYsum[i];
    }
  int *Obs_Count;          	
  Obs_Count = IVECTOR(0, max);
#endif




  /*** compute the goodness-of-fit test for litter data. *********/
  gfit = 0.0;

  for (k = 1; k <= Nobs; k++)
    {
      SR[k] = (Yp[k] - GYpp[k])/sqrt(Gvar[k]);
	  if (!isnan(SR[k]))
        gfit += SR[k]*SR[k];
    }

  zOut->obsChi2 = gfit;
  fprintf (fp_out,
#ifndef RBMDS
          "\n\n\n\n Observed Chi-square = %10.4f\n"
#else
          "\n\n\n\n Observed Chi-square = %30.22\n"
#endif
          , gfit);

  df = Nobs - Model_DF(bounded);
  zOut->df = df;
  fprintf (fp_out, " DF = %d\n", df);

/*** Bootstrap method for nested models  */

  /*** Set up random number generator for GSL */
  gsl_rng_env_setup();
  type = gsl_rng_default;
  r = gsl_rng_alloc(type);    
  gsl_rng_set (r, BSSeed);

  fprintf (fp_out, "\n \t\t  Bootstrapping Results  \n");
  fprintf (fp_out, "\nNumber of Bootstrap Iterations per run: %d\n", BSIter);
  fprintf (fp_out, "\n\t\t    Bootstrap Chi-square Percentiles");
  fprintf (fp_out, "\n Bootstrap");
  fprintf (fp_out, "\n    Run        P-value    50th     90th     95th     99th");
  fprintf (fp_out, "\n----------------------------------------------------------- ");


#ifdef BSDEBUG
  fprintf (fp_bsdbg, "\n\n ================================================================================================ \n");
  fprintf (fp_bsdbg, "\n \t\t\t\t Bootstrapping Debug Log\n");
  fprintf (fp_bsdbg, " ================================================================================================ ");
#endif

  for (l=1; l<=BSLoops; l++)   /*** bootstrap run loop */
    {
    SRsqCount = 0;
  

#ifdef BSDEBUG
    fprintf (fp_bsdbg, "\n ================================================================================================ ");
    fprintf (fp_bsdbg, "\n \t\t\t\t Bootstrapping Loop # %d\n", l);
    fprintf (fp_bsdbg, "\n GSL random number generator type = %s\n", gsl_rng_name (r));
    fprintf (fp_bsdbg, " Seed = %ld\n\n", BSSeed);
    fprintf (fp_bsdbg, " GSL beta distribution used to calculate Prob. Cutoff\n");
    fprintf (fp_bsdbg, " when dose correlation coefficents are non-zero\n");
    fprintf (fp_bsdbg, " ============================================================================================================================================================================================= ");
    fprintf (fp_bsdbg, "\n\n                                                           Litter       Actual     Pseudo      Actual         Pseudo");
    fprintf (fp_bsdbg, "\n    Dose       Est._Prob.      Cut_Prob.     Expected       Size       Observed   Observed       SR             SR             Phi                         a                        b");
    fprintf (fp_bsdbg, "\n ============================================================================================================================================================================================= ");
#endif

    for (i = 1; i <= BSIter; i++)    /*** bootstrap iteration loop */
      {

#ifdef BSDEBUG 
      if (i % BSDEBUG == 0)
        {
        fprintf(fp_bsdbg, "\n\n==================");
        fprintf(fp_bsdbg, "\n= Iteration %4d =", i);
        fprintf(fp_bsdbg, "\n==================\n");    
        }
      for (k=0; k<=max; k++) Obs_Count[k] = 0;
#endif

        GSR_newsqsum[l][i] = 0;   	/*** Set SR squared sum to zero for new calcs */
        for (j = 1; j <= Nobs; j++)	/*** Loop over each line in litter data */
          {
          GYp_new[j] = 0;   // reset pseudo-observed value to zero 
          
          /*** find cut-off probability (cutpoint) */
          if ((P[Xg[j]] == 0) || (P[Xg[j]] == 1))             // Use Est._Prob. for cut-off probability when phi = 0 or 1
            cutpoint = GEp[j];
          else
            {
            // find cut-off probability from beta distribution
			//printf("Xg[j] = %d\n", Xg[j]);
			//printf("P[Xg[j]] = %f\n", P[Xg[j]]);
			if (GEp[j] <= eps)    //Be sure not to evaluate a,b and beta distribution if GEp[j] = 0.  This will skip next if statement also: if (isnan(cutpoint))
		       cutpoint = 0.0;
		    else {
               double tempvalue = GEp[j]*(1.0-GEp[j])/(GEp[j]*(1.0-GEp[j])*P[Xg[j]]) - 1.0;
			//printf("tempvalue = %f\n", tempvalue);
			//printf("GEp[j] = %f\n", GEp[j]);
               a = GEp[j]*tempvalue;
               b = (1.0 - GEp[j])*tempvalue; 
               cutpoint = gsl_ran_beta(r, a, b);
			//printf("cutpoint = %f\n", cutpoint);
			   }
            }

          //printf("a = %f\n", a);
		  //printf("b = %f\n", b);
          /*** use cumulative method if cutpoint = NaN */
          if (isnan(cutpoint))
              {
              cum_beta_comp_low = gsl_cdf_beta_P(0.00001, a, b);  // low end of cumulative scale
              cum_beta_comp_hi = gsl_cdf_beta_P(0.99999, a, b);   // hi end of cumulative scale
              cum_beta = gsl_rng_uniform(r);                  // random value to compare to cumulative scale
              
              if (cum_beta < cum_beta_comp_low)          // if random value is less than low end 
                cutpoint = 0;
              else if (cum_beta >= cum_beta_comp_hi)        // if random value is greater than high end 
                cutpoint = 1.0 - cum_beta_step;
              else if (fabs(cum_beta_comp_low - cum_beta)< fabs(cum_beta_comp_hi - cum_beta))   // if random value is closer to low end 
                {
                cum_beta_comp = 0;
                x = 0;
                while (x<=1.0 && cum_beta > cum_beta_comp)
                  {
                  cutpoint = x;
                  x += cum_beta_step;
                  cum_beta_comp = gsl_cdf_beta_P(x, a, b);
                  }
                }
              else              // if random value is closer to high end
                {
                cum_beta_comp = 1.0;
                x = 1;
                cutpoint = 1.0;
                while (x>=0.0 && cum_beta < cum_beta_comp)
                  {
                     x -= cum_beta_step;
                     cutpoint = x;
                     cum_beta_comp = gsl_cdf_beta_P(x, a, b);
                  }
                }
              }



          /*** Calculate pseudo-observed value */
		  if (GEp[j] <= eps)  // If est.prob = 0, then bootstrap must produce zero responsders for this observation
			  GYp_new[j] = 0;   
          else if (P[Xg[j]] != 1)
            {
            for (k = 1; k <= GYsum[j]; k++)
              {
              urand = gsl_rng_uniform (r);	/*generate uniform random number*/
              if (urand <= cutpoint) GYp_new[j]++;    /* Check random number against cutoff-probability.  */
              }
            }
          else
            {
            /*** phi = 1 method */
            urand = gsl_rng_uniform (r);
            if (urand <= cutpoint)
              GYp_new[j] = GYsum[j];
            else
              GYp_new[j] = 0;
            }


          /*** Calculate Scaled Residual based on pseudo-observation (GYp_new[]) */
          GSR_new[j] = Gvar[j] > 0 ? (GYp_new[j] - GYpp[j])/sqrt(Gvar[j]) : 0;
          GSR_newsqsum[l][i] += SQR(GSR_new[j]);         

#ifdef BSDEBUG
  if (i % BSDEBUG == 0)
    {
    Obs_Count[(int)GYp_new[j]]++; 
    //fprintf(fp_bsdbg, "%9.4f\t%6.3f  \t%6.3f   %5.0f      %7.3f\t%5.0f\t     %5.0f\t%9.4f\t%9.4f\n", Xi[j], GEp[j], cutpoint, GYpp[j], GYsum[j], Yp[j], GYp_new[j], SR[j], GSR_new[j]);
    fprintf(fp_bsdbg, "%9.4f       %7.5f        %8.5f      %7.3f      %7.3f      %5.0f      %5.0f      %9.4f      %9.4f      %18.14f      %18.14f      %18.14f\n", 
                      Xi[j], GEp[j], cutpoint, GYpp[j], GYsum[j], Yp[j], GYp_new[j], SR[j], GSR_new[j], P[Xg[j]], a, b);
    }
#endif
          }  

        if (GSR_newsqsum[l][i] >= gfit)
          SRsqCount++;
    
#ifdef BSDEBUG
  if (i % BSDEBUG == 0)
    {
    fprintf(fp_bsdbg, "\n\t\t\t\t\t\t\t\t\t\t\t       ========	      ========\n");
    fprintf(fp_bsdbg, "\t SRsqCount = %5d\t\t\t\t\t\t  Sum of SR^2:       %9.4f	     %9.4f\n\n", SRsqCount, gfit, GSR_newsqsum[l][i]);
    for(k=0; k<=max; k++)
      {
      fprintf(fp_bsdbg, " %d pseudo-obs of value %d\n", Obs_Count[k], k);
      }
    }
#endif

        }

    pv[l] = (double) SRsqCount/BSIter;	/* compute p-value for each BS loop */

  }
  
  /*** Calculate chi-square percentiles for each BS loop */
  pavg = 0;

  for (l=1; l<=BSLoops; l++)
    {
      bsRow[l-1].bsNum = l;
      bsRow[l-1].pvalue = pv[l];
      fprintf(fp_out, "\n %5d          %6.4f  ", l, pv[l]);
            
      /*** compute percentile values of sum of SR^2 ***/
      qsort (&GSR_newsqsum[l][1], BSIter, sizeof(GSR_newsqsum[1][1]), compare_doubles);
      for (k=0; k<sizeof(percentiles)/sizeof(percentiles[0]); k++)
        {
    
        double temp = percentiles[k] * BSIter;
        if ((ceilf(temp)==temp) && (floorf(temp)==temp))
          {
          perloc = (int)temp;
          fprintf(fp_out, "%7.4f  ", GSR_newsqsum[l][perloc]);
	  pctTemp[k] = GSR_newsqsum[l][perloc];
          }
        else
          {
          perloc = (int)ceilf(temp);
          fprintf(fp_out, "%7.4f  ", (GSR_newsqsum[l][perloc]+GSR_newsqsum[l][perloc+1])/2.0);
	  pctTemp[k] = (GSR_newsqsum[l][perloc]+GSR_newsqsum[l][perloc+1])/2.0;
          }
        }
      bsRow[l-1].pct50 = pctTemp[0];
      bsRow[l-1].pct90 = pctTemp[1];
      bsRow[l-1].pct95 = pctTemp[2];
      bsRow[l-1].pct99 = pctTemp[3];
      
    pavg += pv[l];
    }


  
    /*** combined P-value ***/
    pavg /= 3.0;
    fprintf (fp_out, "\n----------------------------------------------------------- ");
    fprintf (fp_out, "\n Combined       %6.4f  ", pavg);

    /*** combined chi-square percentiles ***/
    qsort (&GSR_newsqsum[1][1], BSLoops*BSIter, sizeof(GSR_newsqsum[1][1]), compare_doubles);
    double *pGSRsum = &GSR_newsqsum[1][1];
    bsRow[3].bsNum = -9999;
    bsRow[3].pvalue = pavg;
    for (k=0; k<sizeof(percentiles)/sizeof(percentiles[0]); k++)
        {      
        double temp = percentiles[k] * BSLoops*BSIter;
        if ((ceilf(temp)==temp) && (floorf(temp)==temp))
          {
          perloc = (int)temp;
          fprintf(fp_out, "%7.4f  ", *(pGSRsum+perloc-1));
	  pctTemp[k] = *(pGSRsum+perloc-1);
          }
        else
          {
          perloc = (int)ceilf(temp);
          fprintf(fp_out, "%7.4f  ", (*(pGSRsum+perloc-1)+*(pGSRsum+perloc))/2.0);
	  pctTemp[k] = (*(pGSRsum+perloc-1)+*(pGSRsum+perloc))/2.0;
          }
        }
    bsRow[3].pct50 = pctTemp[0];
    bsRow[3].pct90 = pctTemp[1];
    bsRow[3].pct95 = pctTemp[2];
    bsRow[3].pct99 = pctTemp[3];

    fprintf(fp_out, "\n\n\n\nThe results for three separate runs are shown.  If the estimated p-values are sufficiently");
    fprintf(fp_out, "\nstable (do not vary considerably from run to run), then then number of iterations is");
    fprintf(fp_out, "\nconsidered adequate.  The p-value that should be reported is the one that combines");
    fprintf(fp_out, "\nthe results of the three runs.  If sufficient stability is not evident (and especially");
    fprintf(fp_out, "\nif the p-values are close to the critical level for determining adequate fit, e.g., 0.05),");
    fprintf(fp_out, "\nthen the user should consider increasing the number of iterations per run.\n");
    
    


#ifdef BSDEBUG
  if (fclose(fp_bsdbg) != 0)
    printf("Error closing bootstrap file");
#endif

/*** End bootstrap method for nested models - Cody Simmons */


  FREE_DVECTOR (GEp, 1, Nobs);
  FREE_DVECTOR (GYsum, 1, Nobs);
  FREE_DVECTOR (Gcount, 1, Nobs);
  FREE_DVECTOR (Gvar, 1, Nobs);
  FREE_IVECTOR (GrpSize, 1, ngrp);
  FREE_IVECTOR (P, 1, nparm);
  gsl_rng_free(r);    /* free memory from random generator */
}

/*************************************************
*SRoI - Used to find the scaled residual of interest
*
**************************************************/
void mySRoI (NDichotomous_Result_t *zOut, int ngrp, int Nobs, double GXi[], double Xi[], int Xg[], double SR[], double LSC[], double meanLSC, double BMD)
{
  int i, j;
  int locDose = 0;	/*location of closest dosegroup */
  int locLSC = 0;       /*location of closest LSC(first found) */
  int litSR = 1;    	/*number of litters with closest LSC*/
  double maxSR, minSR, avgSR, maxabsSR, minabsSR, avgabsSR;
  double closeLSC = 0;
  double closeDose = 0;
  double idiff, diff;


  /*** choose dose group closest to BMD */
  diff = DBL_MAX;
  for (i=1; i<=ngrp; i++)
    {
    idiff = fabs(BMD - GXi[i]);
    if (idiff < diff)
      {
      diff = idiff;
      locDose = i;
      closeDose = GXi[i];
      }
    }


  /*** choose LSC closest to mean LSC */
  diff = DBL_MAX;
  for (i=1; i<=Nobs; i++)
    {
    if (Xi[i] == closeDose)  // picks out dose group that is closest to BMD
      {
      idiff = fabs(LSC[i] - meanLSC);
      if (idiff == diff)
        litSR++;
      if (idiff < diff)
        {
        litSR = 1;             //reset litter count
        diff = idiff;
        closeLSC = LSC[i];     //value of closest LSC
        locLSC = i;        
        }
      }
    }
  

  /*** calculate max, min, average SRoI and |SRoi|*/
  maxSR = SR[locLSC];
  minSR = SR[locLSC];
  avgSR = SR[locLSC];
  maxabsSR = fabs(SR[locLSC]);
  minabsSR = fabs(SR[locLSC]);
  avgabsSR = fabs(SR[locLSC]);

  if (litSR != 1)
    {
    avgSR = 0;
    avgabsSR = 0;
    for (i=locLSC; i<=locLSC+litSR-1; i++)                //relies on dose groups being sorted by LSC
      {
      if (Xi[i] == GXi[locDose] && LSC[i] == closeLSC)
        {
        if (SR[i] > maxSR) maxSR = SR[i];
        if (SR[i] < minSR) minSR = SR[i];
        if (fabs(SR[i]) > maxabsSR) maxabsSR = fabs(SR[i]);
        if (fabs(SR[i]) < minabsSR) minabsSR = fabs(SR[i]);
        avgSR += SR[i];
        avgabsSR += fabs(SR[i]);
        printf("SR[i] = %9.4f, avgabsSR = %9.4f\n", SR[i], avgabsSR);
        }
      }
    avgSR /= litSR;
    avgabsSR /= litSR;
    }

  zOut->srMin = minSR;
  zOut->srAvg = avgSR;
  zOut->srMax = maxSR;
  zOut->srNumLitters = litSR;

  fprintf(fp_out, "\n\n\nScaled Residual(s) for Dose Group Nearest the BMD");
  fprintf(fp_out, "\n------------------------------");


  fprintf(fp_out, "\nMinimum scaled residual for dose group nearest the BMD =      %9.4f", minSR);
  fprintf(fp_out, "\nMinimum ABS(scaled residual) for dose group nearest the BMD = %9.4f", minabsSR);
  fprintf(fp_out, "\nAverage scaled residual for dose group nearest the BMD =      %9.4f", avgSR);
  fprintf(fp_out, "\nAverage ABS(scaled residual) for dose group nearest the BMD = %9.4f", avgabsSR);
  fprintf(fp_out, "\nMaximum scaled residual for dose group nearest the BMD =      %9.4f", maxSR);
  fprintf(fp_out, "\nMaximum ABS(scaled residual) for dose group nearest the BMD = %9.4f", maxabsSR);
  fprintf(fp_out, "\nNumber of litters used for scaled residual for dose group nearest the BMD = %d", litSR);

}

int compare_doubles (const void *a, const void *b)
{
  if(*(double*)a > *(double*)b) return 1;
  else if (*(double*)a < *(double*)b) return -1;
  else return 0;
}
