
//#include <Windows.h>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
//#include "pch.h"
#include "bmds_nested_entry.h"

#define MY_MAX_PARMS 64 // Should be determined dynamically

using namespace std;

int main(int argc, char *argv[])
{
  const int nDS = 1; // Always a single data set
  int n = 0; // Number of rows
  double background = -9999; // Default to estimated
  int useLSC = 1;
  int fixedSize = 1; // Fixed litter size
  int estimateILC = 1; // Whether to estimate phi parms or assume zero
  int BSSeed = 0; // Bootstrap seed value
  vector<BMDSDataRow_t> vInputRows;
  if (argc < 3) {
    cout << "Usage: " << (strrchr(argv[0], '\\') + 1) << " input_file bmr [background]" << endl;
    return (-1);
  }
  if (argc > 3) {
    string strBackground(argv[4]);
    background = stod(strBackground, NULL);
  }
  ifstream input(argv[1]);
  string data;
  std::string::size_type sz;
  string string_bmr(argv[2]);
  NDichotomous_Options_t zOpts;
  zOpts.bmrType = eExtraRisk; // *** HARD CODE risk type for now ***
  zOpts.bmr = stod(string_bmr, &sz);
  zOpts.alpha = 0.05; // *** HARD CODE for now ***
  zOpts.BSIter = 1000;
  zOpts.BSSeed = BSSeed;
  zOpts.bRestrict = 1;
  zOpts.background = background;
  zOpts.useLSC = useLSC;
  zOpts.LSCType = fixedSize;
  zOpts.ILCCalc = estimateILC;


  if (!input) {
    cerr << "Failed to open input file: " << argv[1] << endl;
    return -1; // file failed to open 
  }

  getline(input, data); // Discard header line
  while (std::getline(input, data)) {
    stringstream s_in(data);
    BMDSDataRow_t zTemp;
    getline(s_in, data, ','); // Discard row ident column
    getline(s_in, data, ','); zTemp.dose = stod(data, &sz);
    getline(s_in, data, ','); zTemp.response = stod(data, &sz);
    getline(s_in, data, ','); zTemp.groupSize = stod(data, &sz);
    getline(s_in, data, '\n'); zTemp.col4 = stod(data, &sz);
    vInputRows.push_back(zTemp);
    ++n;
  }

  // Determine the # of dose groups - Assumes the rows are sorted by dose!!!
  int dg = 0; // Number of dose groups
  double dTemp = -1;
  for (auto&& v : vInputRows) {
    if (v.dose != dTemp) { // New dose group
      ++dg;
      dTemp = v.dose;
    } // end if
  } // end for
  BMDSDataSet_t zDataset;
  zDataset.degree = 0; // Not used
  zDataset.rows = &vInputRows[0];
  zDataset.count = n;
  zDataset.dgCount = dg;

  // Allocate storage for the output, including the MLE parms
  NDichotomous_Result_t zResult;
  zResult.PARMS = new double[MY_MAX_PARMS];
  zResult.litterRow = new NDLitterRow_t[n];
  zResult.bsRow = new NDBootstrapRow_t[4];
  zResult.plotData = new NDPlotRow_t[dg];
  int status;
  cout << "Running nested-logistic: riskType= " << zOpts.bmrType << ", bmr= "
       << zOpts.bmr << endl;
  status = nlogistic(&zResult, nDS, &zDataset, &zOpts);
  cout << "nested-logistic run completed with status= " << status << endl;
  if (status == 0) {
    cout << "Results: BMD= " << zResult.BMD << ", BMDL= "
         << zResult.BMDL << ", BMDU= " << zResult.BMDU << endl;
    cout << "MAP= " << zResult.MAP << ", AIC= " << zResult.AIC
         << ", BSSeed= " << zResult.BSSeed << endl;
    cout << "obsChi2= " << zResult.obsChi2 << " df= " << zResult.df
      << "srMin= " << zResult.srMin << " srAvg= " << zResult.srAvg << " srMax= "
      << zResult.srMax << " srLitters= " << zResult.srNumLitters
      << " fixedLSC= " << zResult.fixedLSC << endl;
    for (int i = 0; i < zResult.nparms; i++) {
      cout << "\tParm" << i + 1 << " =" << zResult.PARMS[i] << endl;
    }
    cout << endl;
    for (int i = 0; i < n; i++) {
      cout << zResult.litterRow[i].dose << '\t'
        << zResult.litterRow[i].lsc << '\t'
        << zResult.litterRow[i].estProb << '\t'
        << zResult.litterRow[i].litterSize << '\t'
        << zResult.litterRow[i].expected << '\t'
        << zResult.litterRow[i].observed << '\t'
        << zResult.litterRow[i].scaledResidual << endl;
    }
    cout << endl;
    for (int i = 0; i < 4; i++) {
      cout << zResult.bsRow[i].bsNum
        << '\t' << zResult.bsRow[i].pvalue
        << '\t' << zResult.bsRow[i].pct50
        << '\t' << zResult.bsRow[i].pct90
        << '\t' << zResult.bsRow[i].pct95
        << '\t' << zResult.bsRow[i].pct99
        << endl;
    }
    cout << endl;
    cout << "Plot data:\n";
    cout << "\tDose\tResp\teblower\tepUpper" << endl;
    for (int i = 0; i < dg; i++) {
      cout << '\t' << zResult.plotData[i].dose
        << '\t' << zResult.plotData[i].response
        << '\t' << zResult.plotData[i].ebLower
        << '\t' << zResult.plotData[i].ebUpper
        << endl;
    } // for plotData
  }

  return 0;
}

// Run program: Ctrl + F5 or Debug > Start Without Debugging menu
// Debug program: F5 or Debug > Start Debugging menu

// Tips for Getting Started: 
//   1. Use the Solution Explorer window to add/manage files
//   2. Use the Team Explorer window to connect to source control
//   3. Use the Output window to see build output and other messages
//   4. Use the Error List window to view errors
//   5. Go to Project > Add New Item to create new code files, or Project > Add Existing Item to add existing code files to the project
//   6. In the future, to open this project again, go to File > Open > Project and select the .sln file
