// Main program for executable to run single models
//

#include "stdafx.h"
#include <Windows.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include "bmds_nc_entry.h"

using namespace std;

void printResults(string sModel, BMD_ANAL zResult);

constexpr auto yesVal = 'Y';

int main(int argc, char *argv[])
{
  int n = 0;
  int whichModel = 1;
  double background = -9999; // Default to estimated
  double alpha = 0.05; // *** HARD CODE for now ***
  double bmr = 0;
  RiskType_t riskType = eExtraRisk;
  BMDSInputType_t inputType = eDich_3; // *** HARD CODE input data type for now ***
  int dg = 0; // Number of dose groups
  vector<BMDSInputData_t> vInputData;
  string sModel; // Store model name
  string sMultistage("Multistage-");
  string data; // Used to process input data

  if (argc < 4) {
    cout << "Usage: " << (strrchr(argv[0], '\\') + 1) << " input_file model# bmr [ riskType [background] ]" << endl;
    cout << "\t" << "     model# - 1=Log-logistic, 2=Gamma, 3=Logistic, 4=Probit, 5=Quantal-linear,\n";
    cout << "\t" << "              6=Log-probit, 7=Weibull, 8=Multistage, 9=D-hill" << endl;
    cout << "\t" << "   riskType - 1=Extra risk, 2=Added risk" << endl;
    return (-1);
  }
  // Grab the required arguments
  ifstream input(argv[1]);    // Arg 1 = input file name
  whichModel = stoi(argv[2]); // Arg 2 = model# to run
  string string_bmr(argv[3]); // Arg 3 = bmr value
  bmr = stod(string_bmr, NULL);
  // Process optional arguments if present
  if (argc > 4) riskType = (RiskType_t) stoi(argv[4]); // Arg 4 = riskType
  if (argc > 5) {
    string strBackground(argv[5]); // Arg 5 = background response value
    background = stod(strBackground, NULL);
  }
  BMDSOptions_t zOpts;
  BMDS_D_Opts1_t zOpts1;
  BMDS_D_Opts2_t zOpts2;
  zOpts.bmrType = zOpts2.bmrType = riskType;
  zOpts.bmr = zOpts1.bmr = bmr;
  zOpts.alpha = zOpts1.alpha = alpha;
  zOpts.background = zOpts1.background = background;
  cout << "sizeof zOpts= " << sizeof(zOpts) << ", sizeof opts1= " << sizeof(zOpts1)
    << ", sizeof opts2= " << sizeof(zOpts2) << endl;
  cout << "sizeof zOpts.bmrType= " << sizeof(zOpts.bmrType)
    << ", sizeof zOpts.bLognormal= " << sizeof(zOpts.bLognormal)
    << " sizeof bool= " << sizeof(bool) << endl << endl;

  if (!input) {
    cerr << "Failed to open input file: " << argv[1] << endl;
    return -1; // file failed to open 
  }

  getline(input, data); // Discard header line
  while (std::getline(input, data)) {
    stringstream s_in(data);
    BMDSInputData_t zTemp;
    getline(s_in, data, ','); // Discard row ident column
    getline(s_in, data, ','); zTemp.dose = stod(data, NULL);
    getline(s_in, data, ','); zTemp.response = stod(data, NULL);
    getline(s_in, data, '\n'); zTemp.groupSize = stod(data, NULL);
    vInputData.push_back(zTemp);
    ++n;
  }
  dg = n;
  cout << "\t  Input data " << endl;
  cout << " Dose\tResponse\tN" << endl;
  for (int i = 0; i < n; i++) {
    cout << " " << vInputData[i].dose << '\t' << vInputData[i].response << '\t' << '\t' << vInputData[i].groupSize << '\n';
  }
  cout << endl;

  //// Scale doses by the largest dose
  //double dMaxDose = 0;
  //for (auto&& v : vInputData) if (v.dose > dMaxDose) dMaxDose = v.dose;
  //for (auto& v : vInputData) v.dose /= dMaxDose;
  //cout << " Doses are scaled by " << dMaxDose << endl;

  // We should read the priors in rather than hard code them!!

  // Basic Priors		
#ifdef PRISET1
  double mean_logistic[2] = { 0.0,0.0 }; double sd_logistic[2] = { 2.0,1.0 };
  double mean_probit[2] = { 0,0.0 }; double sd_probit[2] = { 2.0,1.0 };
  double mean_qlinear[2] = { 0.0,0.5 }; double sd_qlinear[2] = { 2.0,1.0 };

  double mean_loglogit[3] = { 0.0,0.0,log(2) }; double sd_loglogit[3] = { 2.0,1.0,sqrt(0.25) };
  double mean_logprobit[3] = { 0,0.0,log(2) }; double sd_logprobit[3] = { 2.0,1.0,sqrt(0.25) };

  double mean_weibull[3] = { 0,log(2),0.0 }; double sd_weibull[3] = { 2.0,sqrt(0.18),1.0 };

  double mean_mult2[3] = { 0.0,0.0,0.0 }; double sd_mult2[3] = { 2.0,sqrt(0.25),1.0 };

  double mean_hill[4] = { 0.0,4.0,0.0,log(10) }; double sd_hill[4] = { 2,2,sqrt(0.25),sqrt(0.06255) };
#endif // PRISET1

  // Diffuse Basic Priors		
#ifdef PRISET2
  double mean_logistic[2] = { 0,0.0 }; double sd_logistic[2] = { 2.0,2.0 };
  double mean_probit[2] = { 0,0.0 }; double sd_probit[2] = { 2.0,2.0 };
  double mean_qlinear[2] = { 0.0,0.5 }; double sd_qlinear[2] = { 2.0,2.0 };

  double mean_loglogit[3] = { 0.0,0.0,log(2) }; double sd_loglogit[3] = { 2.0,1.0,2 * sqrt(0.25) };
  double mean_logprobit[3] = { 0,0.0,log(2) }; double sd_logprobit[3] = { 2.0,1.0,2 * sqrt(0.25) };

  double mean_weibull[3] = { 0,log(2),0.0 }; double sd_weibull[3] = { 2.0,2 * sqrt(0.18),2.0 };
  double mean_gamma[3] = { 0,log(2),0.0 };  double sd_gamma[3] = { 2.0,2 * sqrt(0.18),2.0 };

  double mean_mult2[3] = { 0.0,0.0,0.0 }; double sd_mult2[3] = { 2.0,2 * sqrt(0.25),2.0 };

  double mean_hill[4] = { 0.0,4.0,0.0,log(10) }; double sd_hill[4] = { 2,2,2 * sqrt(0.25),sqrt(2 * 0.06255) };
#endif // PRISET2

  // Wheeler Suggested Priors 1 
#ifdef PRISET3
  double mean_logistic[2] = { -2.5,0.0 }; double sd_logistic[2] = { 1.0,2.0 };
  double mean_probit[2] = { -2.5,0.0 }; double sd_probit[2] = { 1.0,2.0 };
  double mean_qlinear[2] = { -2.5,0.0 }; double sd_qlinear[2] = { 1.0,2.0 };

  double mean_loglogit[3] = { -2.5,0.0,log(2) }; double sd_loglogit[3] = { 1.0,1.0,sqrt(0.2) };
  double mean_logprobit[3] = { -2.5,0.0,log(2) }; double sd_logprobit[3] = { 1.0,1.0,sqrt(0.2) };

  double mean_weibull[3] = { -2.5,log(2),0.0 }; double sd_weibull[3] = { 1.0,sqrt(0.2),2.0 };
  double mean_gamma[3] = { -2.5,log(2),0.0 };  double sd_gamma[3] = { 1.0,sqrt(0.2),2.0 };

  double mean_mult2[3] = { -2.5,0.5,0.0 }; double sd_mult2[3] = { 1.0,sqrt(0.2),2.0 };

  double mean_hill[4] = { -2.5,4.0,0.0,log(4) }; double sd_hill[4] = { 1,2,sqrt(0.5),sqrt(0.2) };
#endif // PRISET3

  // Wheeler Suggested Priors 2
#ifdef PRISET4
  double mean_logistic[2] = { -2.5,0.0 }; double sd_logistic[2] = { 1.0,2.0 };
  double mean_probit[2] = { -2.5,0.0 }; double sd_probit[2] = { 1.0,2.0 };
  double mean_qlinear[2] = { -2.5,-1.0 }; double sd_qlinear[2] = { 1.0,2.0 };

  double mean_loglogit[3] = { -2.5,0.0,log(2) }; double sd_loglogit[3] = { 1.0,1.0,sqrt(0.2) };
  double mean_logprobit[3] = { -2.5,0.0,log(2) }; double sd_logprobit[3] = { 1.0,1.0,sqrt(0.2) };

  double mean_weibull[3] = { -2.5,log(2),0.0 }; double sd_weibull[3] = { 1.0,sqrt(0.2),2.0 };
  double mean_gamma[3] = { -2.5,log(2),0.0 };  double sd_gamma[3] = { 1.0,sqrt(0.2),2.0 };

  double mean_mult2[3] = { -2.5,0.5,0.0 }; double sd_mult2[3] = { 1.0,sqrt(0.2),2.0 };

  double mean_hill[4] = { -2.5,0.0,0.0,log(4) }; double sd_hill[4] = { 1,2,sqrt(0.5),sqrt(0.2) };
#endif // PRISET4

  // Set frequentist priors by number of model parms
  double type_freq_2[2] = { 0.0,0.0 };
  double type_freq_3[3] = { 0.0,0.0,0.0 };
  double type_freq_4[4] = { 0.0,0.0,0.0,0.0 };

#define MY_MAX_PARMS 32 // Should never get close to this many!!!

  BMD_ANAL zResult;
  DichotomousDeviance_t zDev;
  dGoF_t zGoF;
  zGoF.pzRow = new GoFRow_t[n];
  zResult.PARMS = new double[MY_MAX_PARMS];
  zResult.deviance = &zDev;
  zResult.gof = &zGoF;
  zResult.boundedParms = new bool[MY_MAX_PARMS];
  PRIOR aPriors[MY_MAX_PARMS]; // Parameter priors

  // Frequentist: Initialize priors. Specific values are set as needed by model.
  for (int i = 0; i < MY_MAX_PARMS; ++i) {
    aPriors[i].type = 0; // Frequentist
    aPriors[i].initalValue = 0; // Initial mean for prior's distribution
    aPriors[i].stdDev = 1; // Unused for frequentist
    aPriors[i].minValue = -18; // True for most but not all
    aPriors[i].maxValue = 18; // True for all priors
  }

  DModelID_t m_id;
  int retval;
  
  switch (whichModel)
  {
  case 1: // Log-logistic
    m_id = eLogLogistic;
    sModel = "Log-logistic";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.0;
    aPriors[1].initalValue = 0.0; aPriors[1].minValue = 1e-4;
    aPriors[2].initalValue = 0.5; aPriors[2].minValue = 1e-4;
    //bmd_log_logistic(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    //retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    retval = run_dmodel2(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts1, &zOpts2, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 2: // Gamma
    m_id = eGamma;
    sModel = "Gamma";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.009;
    aPriors[1].initalValue = 0.007; aPriors[1].minValue = 1e-4;
    aPriors[2].initalValue = 1; aPriors[2].minValue = 1e-4;
    //bmd_gamma(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 3: // Logistic
    m_id = eLogistic;
    sModel = "Logistic";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.0;
    aPriors[1].initalValue = 0.0; aPriors[1].minValue = 1e-12;
    //bmd_logit(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 4: // Probit
    m_id = eProbit;
    sModel = "Probit";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.0;
    aPriors[1].initalValue = 0.0; aPriors[1].minValue = 1e-12;
    //bmd_probit(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 5: // Quantal Linear
    m_id = eQLinear;
    sModel = "Quantal-linear";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.009;
    aPriors[1].initalValue = 0.01; aPriors[1].minValue = 1e-12;
    //bmd_qlinear(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 6: // Log-probit
    m_id = eLogProbit;
    sModel = "Log-probit";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.0;
    aPriors[1].initalValue = 0.0; aPriors[1].minValue = -18;
    aPriors[2].initalValue = 0.5; aPriors[2].minValue = 1e-4;
    //bmd_log_probit(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 7: // Weibull
    m_id = eWeibull;
    sModel = "Weibull";
    cout << "Running " << sModel << endl;
    aPriors[0].initalValue = 0.009;
    aPriors[1].initalValue = 1.7e-6; aPriors[1].minValue = 1e-6;
    aPriors[2].initalValue = 2.6; aPriors[2].minValue = 1e-6;
    //bmd_weibull(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  case 8: // Multistage
    m_id = eMultistage;
    aPriors[0].initalValue = 0.001;
    aPriors[1].initalValue = 0.001; aPriors[1].minValue = 1e-12;
    aPriors[2].initalValue = 0.001; aPriors[2].minValue = 1e-6;
    for (int i = 2; i < dg; i++) {
      int gap = i - 2; // Number of parm rows beyond the default for degree 2
      int j=3; // Index of next row to add
      while (gap-- > 0) {
        aPriors[j].initalValue = 0.001;
        aPriors[j].minValue = 1e-6;
        ++j;
      } // while
      sModel = sMultistage + to_string(i);
      cout << "Running " << sModel << endl;
      zOpts.degree = i;
      //bmd_multistage(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
      retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
      cout << "model dll status= " << retval << endl;
      // Unscale BMDs
      //zResult.BMD *= dMaxDose; zResult.BMDL *= dMaxDose; zResult.BMDU *= dMaxDose;
      printResults(sModel, zResult);
    } // for
    break;
  case 9: // Hill
    m_id = eDHill;
    sModel = "D-hill";
    cout << "Running " << sModel << endl;
    // Parameters: g, v, a, b
    aPriors[0].initalValue = 1; aPriors[0].minValue = 0; aPriors[0].maxValue = 1;
    aPriors[1].initalValue = 0.001; aPriors[1].minValue = 0; aPriors[1].maxValue = 1;
    aPriors[2].initalValue = -17; aPriors[2].minValue = -20;
    aPriors[3].initalValue = 3.5; aPriors[3].minValue = 1e-4;
    //bmd_hill(&zResult, inputType, &vInputData[0], aPriors, &zOpts, n);
    retval = run_dmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "model dll status= " << retval << endl;
    break;
  default:
    cerr << "Argument 3 is an invalid model code. Valid values are '1' - '9'" << endl;
    exit(1);
    break;
  }
  
  //// Unscale BMDs & print results for models other than multistage
  if (whichModel != 8) {
    //zResult.BMD *= dMaxDose;
    //zResult.BMDL *= dMaxDose;
    //zResult.BMDU *= dMaxDose;
    printResults(sModel, zResult);
  } // endif (whichModel != 8)

  return 0;
}

  void printResults(string sModel, BMD_ANAL zResult)
  {
    // Because of Excel, the DLL must return a unicode string for model_id
#if defined WIN32 || defined _WINDOWS
#define MY_WCOUT wcout
#else
#define MY_WCOUT cout
#endif // WIN32
    DichotomousDeviance_t *zDev = zResult.deviance;
    dGoF_t *zGoF = zResult.gof;

    cout.setf(ios::fixed, ios::floatfield);
    cout.precision(8);
    cout << "Results for model: " << sModel << " version: ";
    MY_WCOUT << zResult.model_id << endl;
    cout << "BMD= " << zResult.BMD << " BMDL= " << zResult.BMDL << " BMDU= "
         << zResult.BMDU << '\n';
    cout << "Full model: ll= " << zDev->llFull << '\n';
    cout << "Fit model: ll= " << zResult.MAP << " dev= " << zDev->devFit << " p-value= "
         << zDev->pvFit << '\n';
    cout << "Reduced model: ll= " << zDev->llReduced << " dev= " << zDev->devReduced
         << " p-value= " << zDev->pvReduced << endl;
    cout << "AIC= " << zResult.AIC << '\n' << endl;
    cout << "parm#" << "\tMLE" << "\tBounded?" << endl;
    for (int i = 0; i < zResult.nparms; i++) {
      cout << i + 1 << '\t' << zResult.PARMS[i];
      if (zResult.boundedParms[i]) cout << '\t' << yesVal;
      cout << endl;
    }
    cout << '\n';
    cout << "Dose " << "EstProb " << "Expected " << "Observed " << "Size "
      << "Sc.Residual" << endl;
    cout.precision(4);
    for (int i = 0; i < zGoF->n; i++) {
      GoFRow_t *zRow = &zGoF->pzRow[i];
      cout << zRow->dose << " " << zRow->estProb << " " << zRow->expected << " " << zRow->observed
        << " " << zRow->size << " " << zRow->scaledResidual << endl;
    }
    cout << "Chi^2= " << zGoF->chiSquare << " df= " << zGoF->df << " p-value= " << zGoF->pvalue << endl;
    return;
  }