// Main program for executable to run single models
//

#include "stdafx.h"
#include <Windows.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include "bmds_cmodel_entry.h"

using namespace std;
void printResults(BMD_C_ANAL zResult, int nRows, double LLConstant);

constexpr auto yesVal = 'Y';

int main(int argc, char *argv[])
{
  int n = 0; // Number of input data rows
  int dg = 0; // Number of dose groups; dg==n for eCont_4 data
  // Define settings with default values (if applicable)
  int whichModel = 1;
  BMRType_t bmrType = eRelativeDev;
  VarType_t myVarType = eConstant;
  bool bLognormal = true;
  bool bSummaryData = true;
  double bmr = 0;
  // Calculate the additive constant value, which is included in BMDS 3.0
  // log-likelihoods but omitted by BMDS 2.x.
  double additiveConstant = 0;


  BMDSInputType_t inputType = eCont_4;
  vector<BMDSInputData_t> vInputData;
  if (argc < 5) {
    cout << "Usage: " << argv[0] << " <model#> <input_file> <bmr type> <bmr value> [Lognormal? [varType [SummaryData?]]]" << endl;
    cout << "\t" << "     model# - 1=all, 2-5=Exp2-5, 6=Hill, 7=Poly, 8=Power" << endl;
    cout << "\t" << "     bmr type - 1=AbsDev, 2=StdDev, 3=RelDev, 4=Point, 6=Hybrid-extra" << endl;
    cout << "\t" << " Lognormal? - 0=No, 1=Yes (default)" << endl;
    cout << "\t" << "    varType - 1=Constance Variance (default), 2=Non-constance variance" << endl;
    cout << "\t" << " SummaryData? - 0=No, 1=Yes (default)" << endl;
    return (-1);
  }
  // Process command line settings
  whichModel = stoi(argv[1]);
  ifstream input(argv[2]);
  bmrType = (BMRType_t) stoi(argv[3]);
  string string_bmr(argv[4]);
  if (argc > 5) {
    bLognormal = (stoi(argv[5]) == 1);
    if (argc > 6) {
      myVarType = (VarType_t)stoi(argv[6]);
      if (argc > 7) bSummaryData = (stoi(argv[7]) == 1);
    } // endif (argc > 6)
  } // endif (argc > 5)

  //std::string::size_type sz;
  bmr = stod(string_bmr, NULL);
  if (!input) {
    cerr << "Failed to open input file: " << argv[1] << endl;
    return -1; // file failed to open 
  }

  string data;
  getline(input, data); // Discard header line
  if (bSummaryData) {
    inputType = eCont_4;
    while (std::getline(input, data)) {
      stringstream s_in(data);
      BMDSInputData_t zTemp;
      getline(s_in, data, ','); // Discard row ident column
      getline(s_in, data, ','); zTemp.dose = stod(data, NULL);
      getline(s_in, data, ','); zTemp.groupSize = stod(data, NULL);
      getline(s_in, data, ','); zTemp.response = stod(data, NULL);
      getline(s_in, data, '\n'); zTemp.col4 = stod(data, NULL); // std. deviation
      vInputData.push_back(zTemp);
      ++n;
    } // end while
    dg = n;

    //additiveConstant = -Y.col(2).sum() * log(2 * M_PI) / 2;
    for (int i = 0; i < n; i++) additiveConstant += vInputData[i].groupSize;
    additiveConstant *= log(2 * M_PI) / 2;

    //// Print out summarized dose-response data
    //cout << "\t  Input data\n" << " Dose\tSize\tMean\tStdDev" << endl;
    //for (int i = 0; i < n; i++) {
    //  cout << " " << vInputData[i].dose
    //    << '\t' << vInputData[i].groupSize
    //    << '\t' << vInputData[i].response
    //    << '\t' << vInputData[i].col4
    //    << '\n';
    //} // end for
  }
  else { // Individual dose/responses
    inputType = eCont_2;
    while (std::getline(input, data)) {
      stringstream s_in(data);
      BMDSInputData_t zTemp;
      getline(s_in, data, ','); // Discard row ident column
      getline(s_in, data, ','); zTemp.dose = stod(data, NULL);
      getline(s_in, data, '\n'); zTemp.response = stod(data, NULL);
      vInputData.push_back(zTemp);
      ++n;
    } // end while
    // Find the number of dose groups
    // In theory the data may not be sorted, but for simplicity assume
    // for now that it is.
    double dTemp = vInputData[0].dose; // Prime the comparison variable
    dg = 1;
    for (int i = 0; i < n; i++) {
      if (vInputData[i].dose != dTemp) {
        dTemp = vInputData[i].dose;
        ++dg;
      } // endif
    } // end for

    //additiveConstant = -X.rows() * log(2 * M_PI) / 2;
    additiveConstant = n * log(2 * M_PI) / 2;

    //// Print out individual dose-response data
    //cout << "\t  Input data\n" << " Dose\tResponse" << endl;
    //for (int i = 0; i < n; i++) {
    //  cout << " " << vInputData[i].dose
    //    << '\t' << vInputData[i].response
    //    << '\n';
    //} // end for
  } // endif bSummaryData

  cout << endl;

  //// Scale doses by the largest dose
  //double dMaxDose = 0;
  //for (auto&& v : vInputData) if (v.dose > dMaxDose) dMaxDose = v.dose;
  //for (auto& v : vInputData) v.dose /= dMaxDose;
  //cout << " Doses are scaled by " << dMaxDose << endl;

#define MY_MAX_PARMS 6

  BMD_C_ANAL zResult;
  zResult.PARMS = new double[MY_MAX_PARMS];
  zResult.gofRow = new cGoFRow_t[n];
  zResult.deviance.llRows = new LLRow_t[NUM_LIKELIHOODS_OF_INTEREST];
  zResult.deviance.testRows = new TestRow_t[NUM_TESTS_OF_INTEREST];
  zResult.boundedParms = new bool[MY_MAX_PARMS];

  PRIOR aPriors[MY_MAX_PARMS]; // Parameter priors
  // Don't worry about priors for now. They are set by the model code.
  memset(aPriors, 0, sizeof(aPriors));

  BMDS_C_Options_t zOpts;
  zOpts.bmrType = bmrType;
  zOpts.alpha = 0.05; // *** HARD CODE for now ***
  zOpts.background = -9999; // *** HARD CODE for now ***
  zOpts.bLognormal = bLognormal;
  zOpts.bmr = bmr;
  zOpts.varType = myVarType;
  zOpts.adverseDirection = 0; // -1;
  zOpts.restriction = -9999; // 0;

  cout << "Analysis settings:\n";
  cout << "BMR: Type= " << bmrType << ", Value= " << bmr << endl;
  cout << "Variance type= " << myVarType << ", bLognormal= " << bLognormal << endl;
  cout << "Adverse direction= " << zOpts.adverseDirection
    << ", Restriction= " << zOpts.restriction << endl << endl;

  CModelID_t m_id;
  int retval;
  switch (whichModel) {
  case 1: // Run all of the models
  case 2: // Do all of the exp models if any one is selected
  case 3:
  case 4:
  case 5:
    cout << "Running Exp2..." << endl;
    zOpts.degree = 2;
    m_id = CModelID_t::eExp2;
    retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "Results for Exp2: status= " << retval << endl;
    printResults(zResult, dg, additiveConstant);
    cout << "Running Exp3..." << endl;
    zOpts.degree++;
    m_id = CModelID_t::eExp3;
    retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "Results for Exp3: status= " << retval << endl;
    printResults(zResult, dg, additiveConstant);
    cout << "Running Exp4..." << endl;
    zOpts.degree++;
    m_id = CModelID_t::eExp4;
    retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "Results for Exp4: status= " << retval << endl;
    printResults(zResult, dg, additiveConstant);
    cout << "Running Exp5..." << endl;
    zOpts.degree++;
    m_id = CModelID_t::eExp5;
    retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "Results for Exp5: status= " << retval << endl;
    printResults(zResult, dg, additiveConstant);
    if (whichModel > 1) break;

  case 6: // Hill
    cout << "Running Hill..." << endl;
    zOpts.degree = 0;
    m_id = CModelID_t::eHill;
    retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "Results for Hill model: status= " << retval << endl;
    printResults(zResult, dg, additiveConstant);
    if (whichModel > 1) break;

  case 7: // Polynomial
    for (int i = 1; i < dg - 1; i++) {
      cout << "Running Polynomial, degree= " << i << endl;
      zOpts.degree = i;
      m_id = CModelID_t::ePoly;
      retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
      cout << "Results for Poly-" << i << " model: status= " << retval << endl;
      printResults(zResult, dg, additiveConstant);
    } // for
    if (whichModel > 1) break;

  case 8: // Power
    cout << "Running Power..." << endl;
    zOpts.degree = 0;
    m_id = CModelID_t::ePow;
    retval = run_cmodel(&m_id, &zResult, &inputType, &vInputData[0], aPriors, &zOpts, &n);
    cout << "Results for Power model: status= " << retval << endl;
    printResults(zResult, dg, additiveConstant);
    break;

  default:
    cerr << "Argument 3 is an invalid model code. Valid values are '10' - '??'" << endl;
    exit(1);
    break;
  } // end switch

  //// Unscale BMDs
  //zResult.BMD *= dMaxDose;
  //zResult.BMDL *= dMaxDose;
  //zResult.BMDU *= dMaxDose;
  
  return 0;
}

void printResults(BMD_C_ANAL zResult, int nRows, double LLConstant)
{
  // Because of Excel, the DLL must return a unicode string for model_id
#if defined WIN32 || defined _WINDOWS
#define MY_WCOUT wcout
#else
#define MY_WCOUT cout
#endif // WIN32

  cout.setf(ios::fixed, ios::floatfield);
  cout.precision(8);
  //cout << " Model version: ";
  //MY_WCOUT << zResult.model_id << endl;
  cout << " BMD= " << zResult.BMD << " BMDL= " << zResult.BMDL ;
  cout << " BMDU= " << zResult.BMDU << " LL= " << zResult.MAP << endl;
  cout << " AIC= " << zResult.AIC << " bAdverseUp= " << zResult.bAdverseUp << endl;
  cout << "parm#" << "\tMLE" << "\t\tBounded?" << endl;
  for (int i = 0; i < zResult.nparms; i++) {
    cout << i + 1 << '\t' << zResult.PARMS[i];
    if (zResult.boundedParms[i]) cout << "\t\t" << yesVal;
    cout << endl;
  }
  cout << "Dose    " << "\tN       " << "\tObsMean " << "\tEstMean " << "\tObs SD  "
    << "\tEst SD  " << "\tScResidual" << endl;
  for (int i = 0; i < nRows; i++) {
    cout << zResult.gofRow[i].dose << '\t' << zResult.gofRow[i].size << '\t'
      << zResult.gofRow[i].obsMean << '\t' << zResult.gofRow[i].estMean << '\t'
      << zResult.gofRow[i].obsStDev << '\t' << zResult.gofRow[i].estStDev
      << '\t' << zResult.gofRow[i].scaledResidual << endl;
  }
  cout << endl;
  cout << "LL constant that BMDS 2.x omits: " << LLConstant << endl;
  cout << "Model#" << "\tLL" << "\t\tnparms" << "\t aic" << "\t\tBMDS2 LL" << endl;
  for (int i = 0; i < NUM_LIKELIHOODS_OF_INTEREST; i++) {
    cout << zResult.deviance.llRows[i].model << '\t' << zResult.deviance.llRows[i].ll
      << '\t' << zResult.deviance.llRows[i].nParms
      << '\t' << zResult.deviance.llRows[i].aic
      << '\t' << zResult.deviance.llRows[i].ll - LLConstant << endl;
  }
  cout << endl;
  cout << "Test#" << "\tdeviance" << "\tdf " << "\tp-value" << endl;
  for (int i = 0; i < NUM_TESTS_OF_INTEREST; i++) {
    cout << zResult.deviance.testRows[i].testNumber
      << '\t' << zResult.deviance.testRows[i].deviance
      << '\t' << zResult.deviance.testRows[i].df
      << '\t' << zResult.deviance.testRows[i].pvalue
      << endl;
  }
  return;
}